from ltk.actions.action import *

class RmAction(Action):
    def __init__(self, path):
        Action.__init__(self, path)
        self.use_delete = False

    def rm_action(self, file_patterns, **kwargs):
        try:
            removed_folder = False
            for pattern in file_patterns:
                if os.path.isdir(pattern):
                    # print("checking folder "+self.norm_path(pattern))
                    if self.folder_manager.folder_exists(self.norm_path(pattern)):
                        self.folder_manager.remove_element(self.norm_path(pattern))
                        logger.info("Removed folder "+pattern)
                        removed_folder = True
                    else:
                        logger.warning("Folder "+str(pattern)+" has not been added and so can not be removed")
            if 'directory' in kwargs and kwargs['directory']:
                if not removed_folder:
                    logger.info("No folders to remove at the given path(s)")
                return
            matched_files = None
            if isinstance(file_patterns,str):
                file_patterns = [file_patterns]
            if 'force' in kwargs and kwargs['force']:
                force = True
            else:
                force = False
            if 'id' in kwargs and kwargs['id']:
                useID = True
            else:
                useID = False
            if 'remote' in kwargs and kwargs['remote']:
                self.use_delete = True
            else:
                self.use_delete = False
            if 'all' in kwargs and kwargs['all']:
                local = False
                self.folder_manager.clear_all()
                removed_folder = True
                logger.info("Removed all folders.")
                useID = False
                matched_files = self.doc_manager.get_file_names()
            elif 'local' in kwargs and kwargs['local']:
                local = True
                if 'name' in kwargs and kwargs['name']:
                    matched_files = []

                    for pattern in file_patterns:
                        doc = self.doc_manager.get_doc_by_prop("name",pattern)
                        if doc:
                            matched_files.append(doc['file_name'])
                else:
                    if len(file_patterns) == 0:
                        self.folder_manager.clear_all()
                        removed_folder = True
                        logger.info("Removed all folders.")
                        useID = False
                        matched_files = self.doc_manager.get_file_names()

            elif not useID:
                local = False
                # use current working directory as root for files instead of project root
                if 'name' in kwargs and kwargs['name']:
                    matched_files = []

                    for pattern in file_patterns:
                        doc = self.doc_manager.get_doc_by_prop("name",pattern)
                        if doc:
                            matched_files.append(doc['file_name'])
                else:
                    matched_files = self.get_doc_filenames_in_path(file_patterns)
            else:
                local = False
                matched_files = file_patterns
            if not matched_files or len(matched_files) == 0:
                if useID:
                    raise exceptions.ResourceNotFound("No documents to remove with the specified id")
                elif removed_folder:
                    logger.info("No documents to remove")
                elif local:
                    raise exceptions.ResourceNotFound("Too many agruments, to specify a document to be removed locally use -l in association with -n")
                elif not 'all' in kwargs or not kwargs['all']:
                    raise exceptions.ResourceNotFound("No documents to remove with the specified file path")
                else:
                    raise exceptions.ResourceNotFound("No documents to remove")
            is_directory = False
            for pattern in file_patterns: # If attemping to remove any directory, don't print failure message
                basename = os.path.basename(pattern)
                if not basename or basename == "":
                    is_directory = True
            for file_name in matched_files:
                # title = os.path.basename(os.path.normpath(file_name)).split('.')[0]
                self._rm_document(self.norm_path(file_name).replace(self.path,""), useID, (force or local))

        except Exception as e:
            # Python 3
            log_error(self.error_file_name, e)
            # End Python 3
            if 'string indices must be integers' in str(e):
                logger.error("Error connecting to Lingotek's TMS")
            else:
                logger.error("Error on remove: "+str(e))

    def _rm_clone(self, file_name):
        trans_files = []
        entry = self.doc_manager.get_doc_by_prop("file_name", file_name)
        if entry:
            if 'locales' in entry and entry['locales']:
                locales = entry['locales']
                for locale_code in locales:
                    if locale_code in self.locale_folders:
                        download_root = self.locale_folders[locale_code]
                    elif self.download_dir and len(self.download_dir):
                        download_root = os.path.join((self.download_dir if self.download_dir and self.download_dir != 'null' else ''),locale_code)
                    else:
                        download_root = locale_code
                    download_root = os.path.join(self.path,download_root)
                    source_file_name = entry['file_name']
                    source_path = os.path.join(self.path,os.path.dirname(source_file_name))

                    trans_files.extend(get_translation_files(file_name, download_root, self.download_option, self.doc_manager))

        return trans_files

    def _rm_document(self, file_name, useID, force):
        try:
            doc = None
            if not useID:
                relative_path = self.norm_path(file_name)
                doc = self.doc_manager.get_doc_by_prop('file_name', relative_path)
                title = os.path.basename(self.norm_path(file_name))
                try:
                    document_id = doc['id']
                except TypeError: # Documents specified by name must be found in the local database to be removed.
                    logger.warning("Document name specified for remove isn't in the local database: {0}".format(relative_path))
                    return
                    # raise exceptions.ResourceNotFound("Document name specified doesn't exist: {0}".format(document_name))
            else:
                document_id = file_name
                doc = self.doc_manager.get_doc_by_prop('id', document_id)
                if doc:
                    file_name = doc['file_name']

            if self.use_delete:
                response = self.api.document_delete(document_id)
            else:
                response = self.api.document_cancel(document_id)
            #print (response)
            if response.status_code != 204 and response.status_code != 202:
                # raise_error(response.json(), "Failed to delete document {0}".format(document_name), True)
                logger.error("Failed to {0} {1} remotely".format('delete' if self.use_delete else 'cancel', file_name))
            else:
                logger.info("{0} has been {1} remotely".format(file_name, 'deleted' if self.use_delete else 'cancelled'))
            if force:
                #delete local translation file(s) for the document being deleted NOTE:this does not seem to be implemented.  The translation files are gathered below but nothing is ever done with them
                trans_files = []
                if 'clone' in self.download_option:
                    trans_files = self._rm_clone(file_name)

                elif 'folder' in self.download_option:
                    trans_files = self._rm_folder(file_name)

                elif 'same' in self.download_option:
                    download_path = self.path
                    trans_files = get_translation_files(file_name, download_path, self.download_option, self.doc_manager)
                self.delete_local(file_name, document_id)
            self.doc_manager.remove_element(document_id)
        except json.decoder.JSONDecodeError:
            logger.error("JSON error on removing document")
        except KeyboardInterrupt:
            raise_error("", "Canceled removing document")
            return
        except Exception as e:
            log_error(self.error_file_name, e)
            logger.error("Error on removing document "+str(file_name)+": "+str(e))

    def _rm_folder(self, file_name):
        trans_files = []
        entry = self.doc_manager.get_doc_by_prop("file_name", file_name)
        if entry:
            if 'locales' in entry and entry['locales']:
                locales = entry['locales']
                for locale_code in locales:
                    if locale_code in self.locale_folders:
                        if self.locale_folders[locale_code] == 'null':
                            logger.warning("Download failed: folder not specified for "+locale_code)
                        else:
                            download_path = self.locale_folders[locale_code]
                    else:
                        download_path = self.download_dir

                    download_path = os.path.join(self.path,download_path)
                    trans_files.extend(get_translation_files(file_name, download_path, self.download_option, self.doc_manager))
        return trans_files



from ltk.actions.action import *
import json
import re
import tempfile
import zipfile

class DownloadAction(Action):
    def __init__(self, path):
        Action.__init__(self, path)
        self.download_root = ''
        self.download_path = ''
        self.DOWNLOAD_NUMBER = 1
        self.default_download_ext = "({0})".format(self.DOWNLOAD_NUMBER)
        self.current_doc = ''
        self.DEFAULT_COMMIT_MESSAGE  = "Translations updated for "
        self.documents_downloaded = False

    def download_by_path(self, file_path, locale_codes, locale_ext, no_ext, auto_format, xliff):
        docs = self.get_docs_in_path(file_path)
        if len(docs) == 0:
            logger.warning("No document found with file path "+str(file_path))
        if docs:
            for entry in docs:
                locales = []
                if locale_codes:
                    locales = locale_codes.split(",")
                elif 'locales' in entry:
                    locales = entry['locales']
                if 'clone' in self.download_option and not locale_ext or (no_ext and not locale_ext):
                    self.download_locales(entry['id'], locales, auto_format, xliff, False)
                else:
                    self.download_locales(entry['id'], locales, auto_format, xliff, True)

    def download_locales(self, document_id, locale_codes, auto_format, xliff, locale_ext=True):
        if locale_codes:
            for locale_code in locale_codes:
                locale_code = locale_code.replace("_","-")
                self.download_action(document_id, locale_code, auto_format, xliff, locale_ext)

    def download_action(self, document_id, locale_code, auto_format, xliff=False, locale_ext=True):
        try:
            latest_document_id = self.get_latest_document_version(document_id) or document_id
            response = self.api.document_content(latest_document_id, locale_code, auto_format, xliff, self.finalized_file)
            entry = None
            entry = self.doc_manager.get_doc_by_prop('id', document_id)
            specific_folder = False
            if entry:
                if self.doc_manager.get_doc_target_folder(entry['file_name']):
                    specific_folder = True
            git_commit_message = self.DEFAULT_COMMIT_MESSAGE
            if response.status_code == 410:
                raise_error(response.json(), "Failed to download " + locale_code + ".", True)
                return
            if response.status_code == 200:
                self.download_path = self.path
                if specific_folder:
                    self.download_path = self.doc_manager.get_doc_target_folder(entry['file_name'])
                elif 'clone' in self.download_option:
                    if not locale_code:
                        print("Cannot download "+str(entry['file_name']+" with no target locale."))
                        return
                    self._clone_download(locale_code)
                else:
                    locale_code = locale_code.replace("-","_")#change to be _ to - to be consistent with the other cases.  Currently the default is xx-XX in all cases except this one (clone off, download folder specified) 
                    if locale_code in self.locale_folders:
                        if self.locale_folders[locale_code] == 'null':
                            logger.warning("Download failed: folder not specified for "+locale_code)
                        else:
                            self.download_path = self.locale_folders[locale_code]
                    else:
                        if 'folder' in self.download_option:
                            self.download_path = self.download_dir
                    locale_code = locale_code.replace("_","-")#changing locale code back to xx-XX form
                if not entry:
                    doc_info = self.api.get_document(document_id)
                    try:
                        file_title = doc_info.json()['properties']['title']
                        title, extension = os.path.splitext(file_title)
                        if not extension:
                            extension = doc_info.json()['properties']['extension']
                            extension = '.' + extension
                        if extension and extension != '.none':
                            title += extension
                    except KeyError as e:
                        log_error(self.error_file_name, e)
                        raise_error(doc_info.json(),
                                    'Something went wrong trying to download document: {0}'.format(document_id), True)
                        return
                    self.download_path = os.path.join(self.download_path, title)
                    logger.info('Downloaded: {0} ({1} - {2})'.format(title, self.get_relative_path(self.download_path), locale_code))
                    self.documents_downloaded = True
                else:
                    file_name = entry['file_name']
                    if not file_name == self.current_doc:
                        self.DOWNLOAD_NUMBER = 1
                        self.current_doc = file_name
                    base_name = os.path.basename(self.norm_path(file_name))
                    if not locale_code:
                        # Don't download source document(s), only download translations
                        logger.info("No target locales for "+file_name+".")
                        return
                    if locale_ext or specific_folder:
                        downloaded_name = self.append_ext_to_file(locale_code, base_name, True)
                    else:
                        downloaded_name = base_name
                    if 'xliff' in response.headers['Content-Type'] and xliff == True:
                        downloaded_name = self.change_file_extension('xlf', downloaded_name)
                    if 'same' in self.download_option and not specific_folder:
                        if self.download_path == self.path:
                            self.download_path = os.path.dirname(file_name)
                        new_path = os.path.join(self.path,os.path.join(self.download_path, downloaded_name))
                        new_locale = downloaded_name.split('.')[1].lower()
                        new_locale = new_locale.replace('_', '-')
                        if not os.path.isfile(new_path) or (locale_code in new_path) or (locale_code.lower() == new_locale):
                            self.download_path = new_path
                        else:
                            self.default_download_ext = "({0})".format(self.DOWNLOAD_NUMBER)
                            downloaded_name = self.append_ext_to_file(self.default_download_ext, base_name, False)
                            self.download_path = os.path.join(self.path,os.path.join(self.download_path, downloaded_name))
                            self.DOWNLOAD_NUMBER += 1
                    else:
                        if self.download_path == "null":
                            self.download_path = os.path.join(self.path,downloaded_name)
                        else:
                            self.download_path = os.path.join(self.path,os.path.join(self.download_path, downloaded_name))
                self.doc_manager.add_element_to_prop(document_id, 'downloaded', locale_code)
                config_file_name, conf_parser = self.init_config_file()

                # create new file and write contents
                try:
                    if response.headers['Content-Type'] == 'application/zip':
                        if self.unzip_file == 'on':
                            self.unzip_finalized_file(response, base_name, locale_code)
                        else:
                            self.download_path = self.download_path + ".zip"
                            downloaded_name = downloaded_name + '.zip'
                            with open(self.download_path, 'wb') as fh:
                                for chunk in response.iter_content(1024):
                                    fh.write(chunk)
                    else:
                        with open(self.download_path, 'wb') as fh:
                            for chunk in response.iter_content(1024):
                                fh.write(chunk)
                    logger.info('Downloaded: {0} ({1} - {2})\n'.format(downloaded_name, self.get_relative_path(self.download_path), locale_code))

                    # configure commit message
                    if (downloaded_name + ": ") not in git_commit_message:
                        if self.documents_downloaded:
                            git_commit_message += '; '
                        git_commit_message += downloaded_name + ": "
                        document_added = True
                        self.documents_downloaded = True
                    if document_added:
                        git_commit_message += locale_code
                    else:
                        git_commit_message += ', ' + locale_code

                except:
                    logger.warning('Error: Download failed at '+self.download_path)
                    return self.download_path

                git_autocommit = conf_parser.get('main', 'git_autocommit')
                if git_autocommit in ['True', 'on']:
                    if self.git_auto.repo_exists(self.download_path) and os.path.isfile(self.download_path):
                        # add, commit, and push to Github

                        self.git_auto.add_file(self.download_path)
                        self.git_auto.commit(git_commit_message)
                        self.git_auto.push()
                        print("\n")

                        # reset documents downloaded
                        self.documents_downloaded = False

                return self.download_path
            else:
                printResponseMessages(response)
                if entry:
                    raise_error(response.json(), 'Failed to download content for {0} ({1})'.format(entry['name'], document_id), True)
                else:
                    raise_error(response.json(), 'Failed to download content for id: {0}'.format(document_id), True)
        except Exception as e:
            log_error(self.error_file_name, e)
            if 'string indices must be integers' in str(e) or 'Expecting value: line 1 column 1' in str(e):
                logger.error("Error connecting to Lingotek's TMS")
            else:
                logger.error("Error on download: "+str(e))

    def _clone_download(self, locale_code):
        try:
            locale_folders = {}
            for key, value in self.locale_folders.items():
                key = key.replace('_', '-')
                locale_folders[key] = value
            if locale_code in locale_folders:
                self.download_root = locale_folders[locale_code]
            elif self.download_dir and len(self.download_dir):
                self.download_root = os.path.join((self.download_dir if self.download_dir and self.download_dir != 'null' else ''),locale_code)
            else:
                self.download_root = locale_code
            self.download_root = os.path.join(self.path, self.download_root)
            self.download_path = self.download_root
            target_dirs = self.download_path.split(os.sep)
            incremental_path = ""
            if not os.path.exists(self.download_root):
                os.mkdir(self.download_root)
                #print("Created directory: "+ download_root)
            if target_dirs:
                for target_dir in target_dirs:
                    incremental_path += target_dir + os.sep
                    #print("target_dir: "+str(incremental_path))
                    new_path = os.path.join(self.path,incremental_path)
                    # print("new path: "+str(new_path))
                    if not os.path.exists(new_path):
                        try:
                            os.mkdir(new_path)
                            # print("Created directory "+str(new_path))
                        except Exception as e:
                            log_error(self.error_file_name, e)
                            logger.warning("Could not create cloned directory "+new_path)
        except IOError as e:
            print(e.errno)
            print(e)

    def change_file_extension(self, new_ext, base_name):
        name_parts = base_name.split('.')
        new_name = None
        if(len(name_parts) > 1):
            name_parts[-1] = new_ext
            new_name = '.'.join(part for part in name_parts)
        else:
            new_name = name_parts[0] + '.' + new_ext
            self.download_path = os.path.join(self.path, os.path.join(self.download_path, new_name))
            # print("New name: {0}".format(new_name))
            # print("Download path: {0}".format(self.download_path))

        return new_name

    def append_ext_to_file(self, to_append, base_name, append_locale):
        name_parts = base_name.split('.')
        base_locale = None
        loc_check = None
        joined_target = None
        if len(name_parts) > 1:
            for p in name_parts:
                # check locale against global locale
                if p.lower() == self.locale.lower():
                    loc_check = self.locale_check(p)
                    name_parts.remove(p)
                # replace locale '_' with '-' and check against global locale
                elif p.replace('_', '-').lower() == self.locale.lower():
                    loc_check = self.locale_check(p)
                    base_locale = p.replace('_','-')
                    name_parts.remove(p)
            if append_locale:
                if base_locale:  
                    new_target = self.source_to_target(loc_check, to_append.split(to_append[2:3]))
                    joined_target = '_'.join(new_target)
                    name_parts.insert(-1, joined_target)
                else:
                    if loc_check:
                        new_target = self.source_to_target(loc_check, to_append.split(to_append[2:3]))
                        joined_target = '-'.join(new_target)
                        name_parts.insert(-1, joined_target)
                    else:
                        name_parts.insert(-1, to_append)
            else:
                name_parts[0] = name_parts[0] + to_append
            downloaded_name = '.'.join(part for part in name_parts)
            return downloaded_name
        else:
            downloaded_name = name_parts[0] + '.' + to_append
            self.download_path = os.path.join(self.path,os.path.join(self.download_path, downloaded_name))
            return downloaded_name
    
    ''' Splits locale into two parts, i.e. ['en', 'US'] '''
    def locale_check(self, loc):
        parts = re.split('[^a-zA-Z]', loc)
        return parts

    ''' Create new target to match client input '''
    def source_to_target(self, source, target):
        new_target = []
        for x in range(0, len(source)):
            if source[x].isupper():
                new_target.append(target[x].upper())
            else:
                new_target.append(target[x].lower())
        return new_target
            
    def unzip_finalized_file(self, response, base_name, locale_code):
        with tempfile.SpooledTemporaryFile(mode='w+b') as temp_zip:
            for chunk in response.iter_content(1024):
                temp_zip.write(chunk)
            temp_zip.seek(0)
            zip_ref = zipfile.ZipFile(temp_zip)
            with open(self.download_path, 'w+b') as fh:
                filenames = zip_ref.namelist()
                fh.write(zip_ref.open(filenames[0]).read())
            zip_ref.close()



# Using the following encoding: utf-8
# Python 2
# from ConfigParser import ConfigParser, NoOptionError
# End Python 2
# Python 3
from configparser import ConfigParser, NoOptionError
# End Python 3
import requests
import ctypes
import ltk.check_connection
import os
import shutil
import fnmatch
import time
import getpass
import itertools
import copy
from ltk import exceptions
from ltk.apicalls import ApiCalls
from ltk.utils import *
from ltk.managers import DocumentManager, FolderManager
from ltk.constants import CONF_DIR, CONF_FN, SYSTEM_FILE, ERROR_FN, METADATA_FIELDS
import json
from ltk.logger import logger
from ltk.git_auto import Git_Auto
from tabulate import tabulate

class Action:
    def __init__(self, path, watch=False, timeout=60):
        self.host = ''
        self.access_token = ''
        self.project_id = ''
        self.project_name = ''
        self.path = path
        self.community_id = ''
        self.workflow_id = ''  # default workflow id; MT phase only
        self.locale = ''
        self.always_check_latest_doc = 'off'
        self.clone_option = 'on'
        self.finalized_file = 'off'
        self.unzip_file = 'on'
        self.auto_format_option = ''
        self.download_option = 'clone'
        self.download_dir = None  # directory where downloaded translation will be stored
        self.watch_locales = set()  # if specified, add these target locales to any files in the watch folder
        self.git_autocommit = None
        self.git_username = ''
        self.git_password = ''
        self.append_option = 'none'
        self.locale_folders = {}
        self.default_metadata = {}
        self.metadata_prompt = False
        self.metadata_fields = METADATA_FIELDS
        if not self._is_initialized():
            raise exceptions.UninitializedError("This project is not initialized. Please run init command.")
        self._initialize_self()
        self.watch = watch
        self.doc_manager = DocumentManager(self.path)
        self.folder_manager = FolderManager(self.path)
        self.timeout = timeout
        self.api = ApiCalls(self.host, self.access_token, self.watch, self.timeout)
        self.git_auto = Git_Auto(self.path)
        self.error_file_name = os.path.join(self.path, CONF_DIR, ERROR_FN)

    def is_hidden_file(self, file_path):
        # todo more robust checking for OSX files that doesn't start with '.'
        name = os.path.abspath(file_path).replace(self.path, "")
        if self.has_hidden_attribute(file_path) or ('Thumbs.db' in file_path) or ('ehthumbs.db' in file_path):
            return True
        while name != "":
            if name.startswith('.') or name.startswith('~') or name == "4913":
                return True
            name = name.split(os.sep)[1:]
            name = (os.sep).join(name)
        return False

    def has_hidden_attribute(self, file_path):
        """ Detects if a file has hidden attributes """
        try:
            # Python 2
            # attrs = ctypes.windll.kernel32.GetFileAttributesW(unicode(file_path))
            # End Python 2
            # Python 3
            attrs = ctypes.windll.kernel32.GetFileAttributesW(str(file_path))
            # End Python 3
            assert attrs != -1
            result = bool(attrs & 2)
        except (AttributeError, AssertionError):
            result = False
        return result

    def append_location(self, name, path_to_file, in_directory=False):
        repo_directory = path_to_file
        path_sep = os.sep
        config_file_name, conf_parser = self.init_config_file()
        if not conf_parser.has_option('main', 'append_option'): self.update_config_file('append_option', 'none', conf_parser, config_file_name, 'Update: Added optional file location appending (ltk config --help)')
        append_option = conf_parser.get('main', 'append_option')
        if not in_directory:
            while repo_directory and repo_directory != "" and not (os.path.isdir(repo_directory + os.sep+".ltk")):
                repo_directory = repo_directory.split(path_sep)[:-1]
                repo_directory = path_sep.join(repo_directory)
            if repo_directory == "" and append_option != 'none':
                logger.warning('Error: File must be contained within an ltk-initialized directory')
                return name
            path_to_file = path_to_file.replace(repo_directory, '', 1).strip(os.sep)
        if append_option == 'none': return name
        elif append_option == 'full': return '{0} ({1})'.format(name, path_to_file.rstrip(name).rstrip(os.sep))
        elif len(append_option) > 5 and append_option[:5] == 'name:':
            folder_name = append_option[5:]
            if folder_name in path_to_file:
                return '{0} ({1})'.format(name, path_to_file[path_to_file.find(folder_name)+len(folder_name):].rstrip(name).strip(os.sep))
            else: return '{0} ({1})'.format(name, path_to_file.rstrip(name).rstrip(os.sep))
        elif len(append_option) > 7 and append_option[:7] == 'number:':
            try: folder_number = int(append_option[7:])
            except ValueError:
                logger.warning('Error: Value after "number" must be an integer')
                return name
            if(folder_number >=0):
                return '{0} ({1})'.format(name, path_sep.join(path_to_file.rstrip(name).rstrip(os.sep).split(path_sep)[(-1*folder_number) if folder_number != 0 else len(path_to_file):]))
            else:
                logger.warning('Error: Value after "number" must be a non-negative integer')
                return name
        else:
            logger.warning('Error: Invalid value listed for append option. Please update; see ltk config --help')

    def add_document(self, file_name, title, doc_metadata={}, **kwargs):
        ''' adds the document to Lingotek cloud and the db '''
        if ltk.check_connection.check_for_connection() == False:
            logger.warning("Cannot connect to network. Documents added to the watch folder will be translated after you reconnect to the network.")
            while ltk.check_connection.check_for_connection() == False:
                time.sleep(15)

        if self.is_hidden_file(file_name):
            return
        try:
            if not 'locale' in kwargs or not kwargs['locale']:
                locale = self.locale
            else:
                locale = kwargs['locale']

            # add document to Lingotek cloud
            response = self.api.add_document(locale, file_name, self.project_id, self.append_location(title, file_name), doc_metadata, **kwargs)
            if response.status_code == 402:
                raise_error(response.json(), "", True)
            elif response.status_code != 202:
                raise_error(response.json(), "Failed to add document {0}\n".format(title), True)
            else:
                title = self.append_location(title, file_name)
                logger.info('Added document {0} with ID {1}\n'.format(title,response.json()['properties']['id']))
                relative_path = self.norm_path(file_name)

                # add document to the db
                if 'download_folder' in kwargs and kwargs['download_folder']:
                    self._add_document(relative_path, title, response.json()['properties']['id'], response.json()['properties']['process_id'], kwargs['download_folder'])
                else:
                    self._add_document(relative_path, title, response.json()['properties']['id'], response.json()['properties']['process_id'])
                if 'translation_locale_code' in kwargs and kwargs['translation_locale_code']:
                    self._update_document(relative_path, None, kwargs['translation_locale_code'])
        except KeyboardInterrupt:
            raise_error("", "Canceled adding document\n")
        except Exception as e:
            log_error(self.error_file_name, e)
            if 'string indices must be integers' in str(e) or 'Expecting value: line 1 column 1' in str(e):
                logger.error("Error connecting to Lingotek's TMS\n")
            else:
                logger.error("Error on adding document \n"+str(file_name)+": "+str(e))

    def _is_initialized(self):
        actual_path = find_conf(self.path)
        if not actual_path:
            return False
        self.path = os.path.join(actual_path, '')
        if not is_initialized(self.path):
            return False
        return True

    def _initialize_self(self):
        config_file_name = os.path.join(self.path, CONF_DIR, CONF_FN)
        conf_parser = ConfigParser()
        conf_parser.read(config_file_name)
        self.host = conf_parser.get('main', 'host')
        self.access_token = conf_parser.get('main', 'access_token')
        self.project_id = conf_parser.get('main', 'project_id')
        self.community_id = conf_parser.get('main', 'community_id')
        self.workflow_id = conf_parser.get('main', 'workflow_id')
        self.locale = conf_parser.get('main', 'default_locale')
        self.always_check_latest_doc = conf_parser.get('main', 'always_check_latest_doc')


        self.locale = self.locale.replace('_','-')
        try:
            if conf_parser.has_option('main', 'auto_format'):
                self.auto_format_option = conf_parser.get('main', 'auto_format')
            else:
                self.update_config_file('auto_format', 'on', conf_parser, config_file_name, "")
            if conf_parser.has_option('main', 'finalized_file'):
                self.finalized_file = conf_parser.get('main', 'finalized_file')
            else:
                self.update_config_file('finalized_file', 'off', conf_parser, config_file_name, "")
            if conf_parser.has_option('main', 'unzip_file'):
                self.unzip_file = conf_parser.get('main', 'unzip_file')
            else:
                self.update_config_file('unzip_file', 'on', conf_parser, config_file_name, "")
            if conf_parser.has_option('main', 'project_name'):
                self.project_name = conf_parser.get('main', 'project_name')
            if conf_parser.has_option('main', 'download_folder'):
                self.download_dir = conf_parser.get('main', 'download_folder')
            else:
                self.download_dir = None
                self.update_config_file('download_folder', json.dumps(self.download_dir), conf_parser, config_file_name, "")
            if conf_parser.has_option('main', 'watch_locales'):
                watch_locales = conf_parser.get('main', 'watch_locales')
                if watch_locales:
                    self.watch_locales = set(watch_locales.split(','))
                else:
                    # there are no watch locales, so set it to an empty set
                    self.watch_locales = set()
            else:
                self.watch_locales = set()
                self.update_config_file('watch_locales', json.dumps(list(self.watch_locales)), conf_parser, config_file_name, "")
            if conf_parser.has_option('main', 'locale_folders'):
                self.locale_folders = json.loads(conf_parser.get('main', 'locale_folders'))
                locale_folders = {}
                #for key, value in self.locale_folders.items():
                #    key = key.replace('_', '-');
                #    locale_folders[key] = value
                #self.locale_folders = locale_folders
            else:
                self.locale_folders = {}
                self.update_config_file('locale_folders', json.dumps(self.locale_folders), conf_parser, config_file_name, "")
            if conf_parser.has_option('main', 'download_option'):
                self.download_option = conf_parser.get('main', 'download_option')
            else:
                self.download_option = 'clone'
                self.update_config_file('download_option', self.download_option, conf_parser, config_file_name, "")
            if conf_parser.has_option('main', 'clone_option'):
                self.clone_option = conf_parser.get('main', 'clone_option')
            else:
                self.clone_option = 'on'
                self.update_config_file('clone_option', self.clone_option, conf_parser, config_file_name, "")
            if conf_parser.has_option('main', 'always_check_latest_doc'):
                self.always_check_latest_doc = conf_parser.get('main', 'always_check_latest_doc')
            else:
                self.always_check_latest_doc = 'off'
                self.update_config_file('always_check_latest_doc', self.always_check_latest_doc, conf_parser, config_file_name, "")
            if conf_parser.has_option('main', 'git_autocommit'):
                self.git_autocommit = conf_parser.get('main', 'git_autocommit')
            else:
                self.git_autocommit = ''
                self.update_config_file('git_autocommit', self.git_autocommit, conf_parser, config_file_name, "")
            if conf_parser.has_option('main', 'git_username'):
                self.git_username = conf_parser.get('main', 'git_username')
            else:
                self.git_username = ''
                self.update_config_file('git_username', self.git_username, conf_parser, config_file_name, "")
            if conf_parser.has_option('main', 'git_password'):
                self.git_password = conf_parser.get('main', 'git_password')
            else:
                self.git_password = ''
                self.update_config_file('git_password', self.git_password, conf_parser, config_file_name, "")
            if conf_parser.has_option('main', 'append_option'):
                self.append_option = conf_parser.get('main', 'append_option')
            else:
                self.append_option = 'none'
                self.update_config_file('append_option', self.append_option, conf_parser, config_file_name, "")
            if conf_parser.has_option('main', 'default_metadata'):
                self.default_metadata = json.loads(conf_parser.get('main', 'default_metadata'))
            else:
                self.default_metadata = {}
                self.update_config_file('default_metadata', json.dumps(self.default_metadata), conf_parser, config_file_name, "")
            if conf_parser.has_option('main', 'metadata_prompt'):
                self.metadata_prompt = (conf_parser.get('main', 'metadata_prompt').lower() == 'on')
            else:
                self.metadata_prompt = False
                self.update_config_file('metadata_prompt', 'off', conf_parser, config_file_name, "")
            if conf_parser.has_option('main', 'metadata_fields'):
                self.metadata_fields = json.loads(conf_parser.get('main', 'metadata_fields'))
            else:
                self.metadata_fields = METADATA_FIELDS
                self.update_config_file('metadata_fields', json.dumps(self.metadata_fields), conf_parser, config_file_name, "")

        except NoOptionError as e:
            if not self.project_name:
                self.api = ApiCalls(self.host, self.access_token)
                project_info = self.api.get_project_info(self.community_id)
                self.project_name = project_info[self.project_id]
                config_file_name, conf_parser = self.init_config_file()
                log_info = 'Updated project name'
                self.update_config_file('project_name', self.project_name, conf_parser, config_file_name, log_info)

    def _add_document(self, file_name, title, doc_id, process_id, dl_folder=''):
        """ adds a document to db """
        now = time.time()
        # doc_id = json['properties']['id']
        full_path = os.path.join(self.path, file_name)
        last_modified = os.stat(full_path).st_mtime
        if dl_folder:
            dl_folder = os.path.relpath(dl_folder, self.path)
        self.doc_manager.add_document(title, now, doc_id, last_modified, now, file_name, process_id, dl_folder)

    def _update_document(self, file_name, next_document_id=None, locales=None):
        """ updates a document in the db """
        now = time.time()
        file_path = os.path.join(self.path, file_name)
        # sys_last_modified = os.stat(file_name).st_mtime
        sys_last_modified = os.stat(file_path).st_mtime
        entry = self.doc_manager.get_doc_by_prop('file_name', file_name)
        doc_id = entry['id']
        self.doc_manager.update_document('last_mod', now, doc_id)
        self.doc_manager.update_document('sys_last_mod', sys_last_modified, doc_id)
        # whenever a document is updated, it should have new translations
        self.doc_manager.update_document('downloaded', [], doc_id)
        if next_document_id:
            self.doc_manager.update_document('id', next_document_id, doc_id)
            doc_id = next_document_id
        if locales:
            self.doc_manager.update_document('locales', locales, doc_id)

    def locked_doc_response_manager(self, response, document_id, *args, **kwargs):
        if response.status_code == 423 and 'next_document_id' in response.json():
            self.doc_manager.update_document('id', response.json()['next_document_id'], document_id)
            return self.api.document_update(response.json()['next_document_id'], *args, **kwargs), response.json()['next_document_id']
        return response, document_id

    def close(self):
        self.doc_manager.close_db()

    def open(self):
        self.doc_manager.open_db()

    def init_config_file(self):
        config_file_name = os.path.join(self.path, CONF_DIR, CONF_FN)
        conf_parser = ConfigParser()
        conf_parser.read(config_file_name)
        return config_file_name, conf_parser

    def update_config_file(self, option, value, conf_parser, config_file_name, log_info):
        try:
            conf_parser.set('main', option, value)
            with open(config_file_name, 'w') as new_file:
                conf_parser.write(new_file)
            self._initialize_self()
            if (len(log_info)):
                logger.info(log_info+"\n")
        except IOError as e:
            print(e.errno)
            print(e)

    def metadata_wizard(self, set_defaults=False):
        import re
        if set_defaults:
            fields = METADATA_FIELDS
            new_metadata = {}
            prompt_message = "Default Value: "
        else:
            fields = self.metadata_fields
            new_metadata = copy.deepcopy(self.default_metadata) 
            prompt_message = "Value: "
            if all (field in self.default_metadata for field in fields):
                print("All fields have default metadata already set")
                for field in fields:
                    print(field,": ",self.default_metadata[field])
                return self.default_metadata
        for field in fields:
            print("\n===",field,"===")
            if field in self.default_metadata and self.default_metadata[field]:
                if set_defaults:
                    print("Current "+prompt_message,self.default_metadata[field])
                    if not yes_no_prompt("Would you like to change the default value for this field?", default_yes=False):
                        new_metadata[field] = self.default_metadata[field]
                        continue
                else:
                    print(prompt_message,self.default_metadata[field])
                    continue
            # Python 2
            # new_value = raw_input(prompt_message)
            # End Python 2
            # Python 3
            new_value = input(prompt_message)
            #End Python 3
            if not new_value:
                continue
            #validate campaign rating field, which is a number field with a maximum of seven digits and allows positive and negative numbers but no decimals
            if field == "campaign_rating":
                while not re.fullmatch('-?0*[0-9]{1,7}', new_value):
                    print("Value must be an integer between -9999999 and 9999999")
                    new_value = input(prompt_message)
                    #allow blank value to not set/change the field in defaults
                    if not new_value:
                        break
                if not new_value: #check if value was unset
                    continue
                #catch -0 and convert it to 0
                if re.fullmatch('-0+', new_value):
                    new_value = "0"
            #validate require review field, which is either true or false
            elif field == "require_review":
                while new_value.upper() != "TRUE" and new_value.upper() != "FALSE":
                    print("Value must be either TRUE or FALSE")
                    new_value = input(prompt_message)
                    #allow blank value to not set/change the field in defaults
                    if not new_value:
                        break
                if not new_value: #check if value was unset
                    continue
            new_metadata[field] = new_value
        return new_metadata

    def validate_metadata_fields(self, field_options):
        if field_options.lower() == 'all' or field_options == '':
            return True, METADATA_FIELDS
        else:
            converted = field_options.replace(", ",",") #allows for a comma-separated list with or without a single space after commas
            options = converted.split(",")
            for option in options:
                if option not in METADATA_FIELDS:
                    logger.warning("Error: {0} is not a valid metadata field".format(option))
                    return False, None
            return True, options

    def get_relative_path(self, path):
        return get_relative_path(self.path, path)

    def get_current_path(self, path):
        cwd = os.getcwd()
        if cwd in path:
            path = path.replace(cwd,"")
            return path
        else:
            cwd_relative_path = cwd.replace(self.path,"")
            return path.replace(cwd_relative_path+os.sep,"")

    def get_current_abs(self, path):
        # print("orig path: "+str(path))
        cwd = os.getcwd()
        if cwd in path:
            path = path.replace(cwd,"")
        else:
            # print("cwd: "+cwd)
            # print("self.path: "+self.path)
            cwd_relative_path = cwd.replace(self.path,"")
            # print("cwd relative path: "+cwd_relative_path)
            cwd_path = path.replace(cwd_relative_path+os.sep,"")
            # print("cwd path: "+cwd_path)
            path = cwd_path
        # print("current path: "+path)
        # print("abs path: "+os.path.abspath(path))
        return os.path.abspath(path)

    def norm_path(self, file_location):
        # print("original path: "+str(file_location))
        if file_location:
            file_location = os.path.normpath(file_location)
            # abspath=os.path.abspath(file_location)
            # print("abspath: "+str(os.path.abspath(os.path.expanduser(file_location))))
            # print("self.path: "+self.path)
            # print("cwd: "+str(os.getcwd()))
            norm_path = os.path.abspath(os.path.expanduser(file_location)).replace(self.path, '')
            # print("normalized path: "+norm_path)
            # print("joined path: "+str(os.path.join(self.path,file_location)))
            # if file_location == ".." and self.path.rstrip('/') in norm_path:
            #     return norm_path.replace(self.path.rstrip('/'), '')
            if file_location is not "." and ".." not in file_location and os.path.exists(os.path.join(self.path,file_location)):
                # print("returning original path: "+str(file_location))
                return file_location.replace(self.path, '')
            elif ".." in file_location and file_location != "..":
                # print("returning norm path: "+norm_path)
                return norm_path.replace(self.path,'')
            if not os.path.exists(os.path.join(self.path,norm_path)) and os.path.exists(os.path.join(self.path,file_location)):
                # print("Starting path at project directory: "+file_location.replace(self.path, ''))
                return os.path.abspath(os.path.expanduser(file_location.replace(self.path, ''))).replace(self.path, '')
            elif file_location == "..":
                return os.path.abspath(os.path.expanduser(file_location.replace(self.path, ''))).replace(self.path, '')
            return norm_path
        else:
            return None

    def get_docs_in_path(self, path):
        files = get_files(path)
        db_files = self.doc_manager.get_file_names()
        docs = []
        if files:
            for file in files:
                file_name = self.norm_path(file)
                if file_name in db_files:
                    docs.append(self.doc_manager.get_doc_by_prop('file_name',file_name))
        return docs

    def get_doc_filenames_in_path(self, path):
        files = get_files(path)
        db_files = self.doc_manager.get_file_names()
        docs = []
        if files:
            for file in files:
                file_name = self.norm_path(file)
                if file_name in db_files:
                    docs.append(file_name)
        return docs

    def get_doc_locales(self, doc_id, doc_name):
        locales = []
        response = self.api.document_translation_status(doc_id)
        if response.status_code != 200:
            if check_response(response) and response.json()['messages'] and 'No translations exist' in response.json()['messages'][0]:
                return locales
            if doc_name:
                raise_error(response.json(), 'Failed to check target locales for document '+doc_name, True, doc_id)
            else:
                raise_error(response.json(), 'Failed to check target locales for document '+doc_id, True, doc_id)

        try:
            if 'entities' in response.json():
                for entry in response.json()['entities']:
                    locales.append(entry['properties']['locale_code'])
        except KeyError as e:
            print("Error listing translations")
            return
            # return detailed_status
        return locales

    def is_locale_folder_taken(self, new_locale, path):
        # Python 2
        # for locale, folder in self.locale_folders.iteritems():
        # End Python 2
        # Python 3
        for locale, folder in self.locale_folders.items():
        # End Python 3
            if path == folder and not locale == new_locale:
                return locale
        return False

    def get_latest_document_version(self, document_id):
        if self.always_check_latest_doc == 'off':
            return False
        try:
            response = self.api.get_latest_document(document_id)
            if response.status_code != 200:
                print('Latest document was not found')
                return False
            else:
                latest_id = response.json()['properties']['id']
                return latest_id
        except Exception as e:
            log_error(self.error_file_name, e)
            logger.error('Error on getting latest document')
            return False

    def update_document_action(self, file_name, title=None, **kwargs):
        try:
            relative_path = self.norm_path(file_name)
            entry = self.doc_manager.get_doc_by_prop('file_name', relative_path)
            try:
                document_id = entry['id']
            except TypeError as e:
                log_error(self.error_file_name, e)
                logger.error("Document name specified for update doesn't exist: {0}".format(title))
                return
            if title:
                response, previous_doc_id = self.locked_doc_response_manager(self.api.document_update(document_id, file_name, title=title, **kwargs), document_id, file_name, title=title, **kwargs)
            else:
                response, previous_doc_id = self.locked_doc_response_manager(self.api.document_update(document_id, file_name), document_id, file_name, **kwargs)

            if response.status_code == 410:
                target_locales = entry['locales']
                self.doc_manager.remove_element(previous_doc_id)
                print('Document has been archived. Reuploading...')
                self.add_document(file_name, title, self.default_metadata, translation_locale_code=target_locales)
                return True
            elif response.status_code == 402:
                raise_error(response.json(), "Community has been disabled. Please contact support@lingotek.com to re-enable your community", True)
            elif response.status_code == 202:
                try:
                    next_document_id = response.json()['next_document_id']
                except Exception:
                    next_document_id = None
                finally:
                    self._update_document(relative_path, next_document_id)
                    return True
            else:
                raise_error(response.json(), "Failed to update document {0}".format(file_name), True)
                return False
        except Exception as e:
            log_error(self.error_file_name, e)
            if 'string indices must be integers' in str(e) or 'Expecting value: line 1 column 1' in str(e):
                logger.error("Error connecting to Lingotek's TMS")
            else:
                logger.error("Error on updating document"+str(file_name)+": "+str(e))
            return False

    def _target_action_db(self, to_delete, locales, document_id):
        locale_set = set()
        for locale in locales:
            locale_set.add(locale.replace("-", "_"))
        if to_delete:
            curr_locales = self.doc_manager.get_doc_by_prop('id', document_id)['locales']
            updated_locales = set(curr_locales) - locale_set
            self.doc_manager.update_document('locales', updated_locales, document_id)
        else:
            self.doc_manager.update_document('locales', list(locales), document_id)

    def update_doc_locales(self, document_id, include_cancelled=False):
        try:
            locale_map = self.import_locale_info(document_id, include_cancelled)
            locale_info = list(iter(locale_map))
        except exceptions.RequestFailedError as e:
            log_error(self.error_file_name, e)
            locale_info = []
        self.doc_manager.update_document('locales', locale_info, document_id)

    def added_folder_of_file(self, file_path):
        folders = self.folder_manager.get_file_names()
        if not folders:
            #print("not folders")
            return
        for folder in folders:
            folder = os.path.join(self.path, folder)
            if folder in file_path:
                return folder

    def get_new_name(self, file_name, curr_path):
        i = 1
        file_path = os.path.join(curr_path, file_name)
        name, extension = os.path.splitext(file_name)
        while os.path.isfile(file_path):
            new_name = '{name}({i}){ext}'.format(name=name, i=i, ext=extension)
            file_path = os.path.join(curr_path, new_name)
            i += 1
        return file_path

    def import_locale_info(self, document_id, poll=False, include_cancelled=False):
        locale_progress = {}
        response = self.api.document_translation_status(document_id)
        if response.status_code != 200:
            if poll or response.status_code == 404:
                return {}
            else:
                # raise_error(response.json(), 'Failed to get locale details of document', True)
                raise exceptions.RequestFailedError('Failed to get locale details of document')
        try:
            for entry in response.json()['entities']:
                curr_locale = entry['properties']['locale_code']
                curr_progress = int(entry['properties']['percent_complete'])
                curr_status = entry['properties']['status']
                curr_locale = curr_locale.replace('-', '_')
                if include_cancelled or (curr_status.upper() != 'CANCELLED'):
                    locale_progress[curr_locale] = curr_progress
        except KeyError:
            pass
        return locale_progress

    def delete_local(self, title, document_id, message=None):
        # print('local delete:', title, document_id)
        if not title:
            title = document_id
        message = '{0} has been deleted locally'.format(title) if not message else message
        try:
            file_name = self.doc_manager.get_doc_by_prop('id', document_id)['file_name']
        except TypeError:
            logger.info('Document to remove not found in the local database')
            return
        try:
            os.remove(os.path.join(self.path, file_name))
            logger.info(message)
        except OSError:
            logger.info('Something went wrong trying to delete the local file')

    def delete_local_translation(self, file_name):
        try:
            if not file_name:
                logger.info('Please provide a valid file name')

            logger.info('{0} (local translation) has been deleted'.format(self.get_relative_path(file_name)))

            os.remove(os.path.join(self.path, file_name))

        except OSError:
            logger.info('Something went wrong trying to download the local translation')

    def delete_local_path(self, path, message=None):
        path = self.norm_path(path)
        message = '{0} has been deleted locally.'.format(path) if not message else message
        try:
            os.remove(path)
            logger.info(message)
        except OSError:
            logger.info('Something went wrong trying to delete the local file')

def is_initialized(project_path):
    ltk_path = os.path.join(project_path, CONF_DIR)
    if os.path.isdir(ltk_path) and os.path.isfile(os.path.join(ltk_path, CONF_FN)) and \
            os.stat(os.path.join(ltk_path, CONF_FN)).st_size:
        return True
    return False

def choice_mapper(info):
    mapper = {}
    import operator

    sorted_info = sorted(info.items(), key = operator.itemgetter(1))

    index = 0
    for entry in sorted_info:
        if entry[0] and entry[1]:
            mapper[index] = {entry[0]: entry[1]}
            index += 1
    table = []
    for k,v in mapper.items():
        try:
            headers=["ID","Name","UUID"]
            for values in v:
                table.append([ k, v[values], values ])
        except UnicodeEncodeError:
            continue
    print(tabulate(table, headers=headers))
    return mapper

def find_conf(curr_path):
    """
    check if the conf folder exists in current directory's parent directories
    """
    if os.path.isdir(os.path.join(curr_path, CONF_DIR)):
        return curr_path
    elif curr_path == os.path.abspath(os.sep):
        return None
    else:
        return find_conf(os.path.abspath(os.path.join(curr_path, os.pardir)))

def printResponseMessages(response):
    for message in response.json()['messages']:
        logger.info(message)

def get_files(patterns):
    """ gets all files matching pattern from root
        pattern supports any unix shell-style wildcards (not same as RE) """

    cwd = os.getcwd()
    if isinstance(patterns,str):
        patterns = [patterns]
    allPatterns = []
    if isinstance(patterns,list) or isinstance(patterns,tuple):
        for pattern in patterns:
            basename = os.path.basename(pattern)
            if basename and basename != "":
                allPatterns.extend(getRegexFiles(pattern,cwd))
            else:
                allPatterns.append(pattern)
    else:
        basename = os.path.basename(patterns)
        if basename and basename != "":
            allPatterns.extend(getRegexFiles(patterns,cwd))
        else:
            allPatterns.append(patterns)
    matched_files = []
    # print("all patterns: "+str(allPatterns))
    for pattern in allPatterns:
        path = os.path.abspath(pattern)
        # print("looking at path "+str(path))
        # check if pattern contains subdirectory
        if os.path.exists(path):
            if os.path.isdir(path):
                for root, subdirs, files in os.walk(path):
                    # split_path = root.split(os.sep)
                    # print("split_path: {0}".format(split_path))
                    for file in files:
                        if not (("desktop.ini" in file) or ('Thumbs.db' in file) or ('ehthumbs.db' in file)):   # don't add desktop.ini, Thumbs.db, or ehthumbs.db files
                            matched_files.append(os.path.join(root, file))
            else:
                matched_files.append(path)
        # else:
        #     logger.info("File not found: "+pattern)
        # subdir_pat, fn_pat = os.path.split(pattern)
        # if not subdir_pat:
        #     for path, subdirs, files in os.walk(root):
        #         for fn in fnmatch.filter(files, pattern):
        #             matched_files.append(os.path.join(path, fn))
        # else:
        #     for path, subdirs, files in os.walk(root):
        #         # print os.path.split(path)
        #         # subdir = os.path.split(path)[1]  # get current subdir
        #         search_root = os.path.join(root, '')
        #         subdir = path.replace(search_root, '')
        #         # print subdir, subdir_pat
        #         if fnmatch.fnmatch(subdir, subdir_pat):
        #             for fn in fnmatch.filter(files, fn_pat):
        #                 matched_files.append(os.path.join(path, fn))
    if len(matched_files) == 0:
        return None
    return matched_files

def getRegexFiles(pattern,path):
    dir_name = os.path.dirname(pattern)
    if dir_name:
        path = os.path.join(path,dir_name)
    pattern_name = os.path.basename(pattern)
    # print("path: "+path)
    # print("pattern: "+str(pattern))
    matched_files = []
    if pattern_name and not "*" in pattern:
        return [pattern]
    for path, subdirs, files in os.walk(path):
        for fn in fnmatch.filter(files, pattern):
            matched_files.append(os.path.join(path, fn))
    # print("matched files: "+str(matched_files))
    return matched_files



''' Python Dependencies '''
import click
import ctypes
import logging
import os
import subprocess
import sys

''' Internal Dependencies '''
from ltk import __version__
from ltk.actions import *
from ltk.constants import LOG_FN, CONF_DIR, METADATA_FIELDS
from ltk.exceptions import UninitializedError, ResourceNotFound, RequestFailedError, AlreadyExistsError
from ltk.logger import logger, API_LOG_LEVEL, API_RESPONSE_LOG_LEVEL, CustomFormatter
from ltk.utils import remove_powershell_formatting
from ltk.watch import WatchAction

''' Globals '''
CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])
HIDDEN_ATTRIBUTE = 0x02
python_version = sys.version
# Python 3
# if python_version[0] < '3':
#    print('Python 3 is required to run this version of the Lingotek Filesystem connector.\n\nFor other versions and troubleshooting, see: https://github.com/lingotek/filesystem-connector')
#    exit()
# End Python 3


def abort_if_false(ctx, param, value):
    if not value:
        ctx.abort()

def init_logger(path):
    """
    Initializes logger based on path
    """
    logger.setLevel(logging.DEBUG)
    if not path:
        file_handler = logging.FileHandler(LOG_FN)
    else:
        try:
            file_handler = logging.FileHandler(os.path.join(path, CONF_DIR, LOG_FN))

            # if on Windows system, set directory properties to hidden
            if os.name == 'nt':
                try:
                    subprocess.call(["attrib", "+H", os.path.join(path, CONF_DIR)])
                except Exception as e:
                    logger.error("Error on init: "+str(e))
                # logger.info("On Windows, make .ltk folder hidden")
                # # Python 2
                # # ret = ctypes.windll.kernel32.SetFileAttributesW(unicode(os.path.join(path, CONF_DIR)), HIDDEN_ATTRIBUTE)
                # # End Python 2
                # # Python 3
                # ret = ctypes.windll.kernel32.SetFileAttributesW(os.path.join(path, CONF_DIR), HIDDEN_ATTRIBUTE)
                # # End Python 3
                # if(ret != 1):   # return value of 1 signifies success
                #     pass
        except IOError as e:
            #logger.info(e)
            # todo error check when running init without existing conf dir
            try:
                os.mkdir(os.path.join(path, CONF_DIR))
                # if on Windows system, make directory hidden
                if os.name == 'nt':
                    logger.info("On Windows, make .ltk folder hidden")
                    # Python 2
                    # ret = ctypes.windll.kernel32.SetFileAttributesW(unicode(os.path.join(path, CONF_DIR)), HIDDEN_ATTRIBUTE)
                    # End Python 2
                    # Python 3
                    ret = ctypes.windll.kernel32.SetFileAttributesW(os.path.join(path, CONF_DIR), HIDDEN_ATTRIBUTE)
                    # End Python 3
                    if(ret != 1):   # return value of 1 signifies success
                        pass
            except IOError as e:
                print(e.errno)
                print(e)

            file_handler = logging.FileHandler(os.path.join(path, CONF_DIR, LOG_FN))

    console_handler = logging.StreamHandler(sys.stdout)
    file_handler.setLevel(API_LOG_LEVEL)
    file_handler.setFormatter(logging.Formatter('%(asctime)s  %(levelname)s: %(message)s'))
    if quiet:
        console_handler.setLevel(logging.WARNING)
    elif verbosity:
        if verbosity > 1:
            console_handler.setLevel(API_RESPONSE_LOG_LEVEL)
        else:
            console_handler.setLevel(API_LOG_LEVEL)
    else:
        console_handler.setLevel(logging.INFO)
    custom_formatter = CustomFormatter()
    console_handler.setFormatter(custom_formatter)
    logger.addHandler(file_handler)
    logger.addHandler(console_handler)


def print_log(error):
    """
    Prints the error before logger is initialized
    """
    # if not len(logger.handlers):
    print ('Error: {0}'.format(error))
    sys.exit()
    return


@click.group(context_settings=CONTEXT_SETTINGS)
# Python 2
# @click.version_option(version=__version__, message='%(prog)s version %(version)s (Lingotek Filesystem Connector - Python 2)')
# End Python 2
# Python 3
@click.version_option(version=__version__, message='%(prog)s version %(version)s (Lingotek Filesystem Connector - Python 3)')
# End Python 3
@click.option('-q', 'is_quiet', flag_value=True, help='Will only show warnings')
@click.option('-v', 'verbosity_lvl', count=True, help='Show API calls. Use -vv for API responses.')
def ltk(is_quiet, verbosity_lvl):
    global quiet, verbosity
    quiet = is_quiet
    verbosity = verbosity_lvl


@ltk.command()
@click.option('--access_token', help='Your access token')
@click.option('--host', default='myaccount.lingotek.com', # type=click.Choice(['myaccount.lingotek.com', 'cms.lingotek.com', 'clone.lingotek.com']),
              help='Default environment: myaccount.lingotek.com')
@click.option('--client_id', help='This is an advanced option that should only be used for clients that have been issued a specified client_id for analytics')
@click.option('--path', type=click.Path(exists=True),
              help='The path to the project directory to be initialized; defaults to the current directory')
@click.option('-n', '--project_name', help='The preferred project name, defaults to the current directory name')
@click.option('-w', '--workflow_id', default='c675bd20-0688-11e2-892e-0800200c9a66',
              help='The id of the workflow to use for this project; defaults to machine translate only')
@click.option('-b', '--browser', flag_value=True, help='Launches broswer for Authentication')
@click.option('-l', '--locale', default='en_US', help='The default source locale for the project; defaults to en_US')
@click.option('-d', '--delete', flag_value=True,  # expose_value=False, callback=abort_if_false,
              # prompt='Are you sure you want to delete the current project remotely and re-initialize? '
              #        'Use the -c flag if you only want to change the project.',
              help='Delete the current project remotely and re-initialize')
# todo add a 'change' option so don't delete remote project
# @click.option('-c', '--change', flag_value=True, help='Change the Lingotek project. ')
@click.option('--reset', flag_value=True, help='Reauthorize and reset any stored access tokens')
def init(host, access_token, client_id, path, project_name, workflow_id, locale, browser, delete, reset):
    """ Connects a local project to Lingotek """
    try:
        host = 'https://' + host
        if not path:
            path = os.getcwd()
        if not project_name:
            project_name = os.path.basename(os.path.normpath(path))
        init_logger(path)

        init = init_action.InitAction(os.getcwd())
        init.init_action(host, access_token, client_id, path, project_name, workflow_id, locale, browser, delete, reset)

        if(init.turn_clone_on == False):
            # set the download option in config
            config = config_action.ConfigAction(os.getcwd())
            config.set_clone_option('off', print_info=False)

    except (ResourceNotFound, RequestFailedError) as e:
        print_log(e)
        logger.error(e)
        return


@ltk.command()
#TO-DO: @click.option('-a', '--all', help='List all configuration settings (including access token)')
@click.option('-l', '--locale', help='Change the default source locale for the project')
@click.option('-w', '--workflow_id', help='Change the default workflow id for the project')
@click.option('-ld', '--latest_document', help='Toggle always get latest document option \'on\' and \'off\'')
@click.option('-c', '--clone_option', help='Toggle clone download option \'on\' and \'off\'. Turning clone \'on\': Translations will be downloaded to a cloned folder structure, where the root folder for each locale is the locale folder specified in config or a locale folder inside of the default download folder. If a default download folder is not set, then translations will be downloaded to the directory where the project was initialized.' +
                                                'Turning clone \'off\': Downloaded translations will go into their locale folder (if specified) or default download folder, but not in a cloned folder structure. If no default download folder or locale folder is specified, downloaded translations will go to the same folder as their corresponding source files.')
@click.option('-ff', '--finalized_file', help='Toggle finalized file download option \'on\' and \'off\'. Turning finalized file on downloads the finalized file instead of the raw translation.\
                                                A finalized file is typically a file that has undergone some sort of post editing like Desktop Publishing after the translation has compeleted.')
@click.option('-u', '--unzip_file', help='Toggle finalized file UNZIP option \'on\' and \'off\'. With this option \'on\' contents of the finalized file will be placed in the expected directory.')
@click.option('-d', '--download_folder',
              help='Specify a default folder for where downloaded translations should go. Use --none to remove the download folder. Using --none will cause downloaded translations to download to the same folder as their corresponding source files.')
@click.option('-t', '--target_locales', multiple=True,
              help='Specify target locales that documents in watch_folder should be assigned; may either specify '
                   'with multiple -t flags (ex: -t locale -t locale) or give a list separated by commas and no spaces '
                   '(ex: -t locale,locale)')

@click.option('-p', '--locale_folder', nargs=2, type=str, multiple=True, help='For a specific locale, specify the root folder where downloaded translations should appear. Use --none for the path to clear the download folder for a specific locale. Example: -p fr_FR translations/fr_FR. Note: This only works with clone option \'on\'')
@click.option('-r', '--remove_locales', flag_value=True, help='Remove all locale folders and use the default download location instead.')
@click.option('-g', '--git', help='Toggle Git auto-commit option on and off')
@click.option('-gu', '--git_credentials', is_flag=True, help='Open prompt for Git credentials for auto-fill (\'none\' to unset); only enabled for Mac and Linux')
@click.option('-a', '--append_option', help='Change the format of the default name given to documents on the Lingotek system.  Define file information to append to document names as none, full, number:+a number of folders down to include (e.g. number:2), or name:+a name of a directory to start after if found in file path (e.g. name:dir). Default option is none.')
@click.option('-f', '--auto_format', help='Toggle auto format option \'on\' and \'off\'. Applies formatting during download.')
@click.option('-md', '--metadata_defaults', is_flag=True, help='Launch the metadata wizard to set the default metadata that will be sent with every document.')
@click.option('-mp', '--metadata_prompt', help='Toggle prompting for metadata with every add and push.  Use the argument \'on\' to enable this prompt or \'off\' to disable it.')
@click.option('-mf', '--metadata_fields', help="Set the fields that the metadata wizard will use when adding or pushing documents.  All default metadata will still be sent.  Enter the fields to prompt for as a comma-separated list with no spaces, or enter 'all' to include all the fields.  Valid fields are: "+', '.join(str(field) for field in METADATA_FIELDS))

def config(**kwargs):
    """ View or change local configuration """
    try:
        action = config_action.ConfigAction(os.getcwd())
        init_logger(action.path)
        for f in kwargs:
            if kwargs[f]:
                temp = remove_powershell_formatting(kwargs[f])
                kwargs[f] = temp
        action.config_action(**kwargs)
    except (UninitializedError, RequestFailedError) as e:
        print_log(e)
        logger.error(e)
        return


@ltk.command(short_help="Add files and folders")
@click.argument('file_names', required=True, nargs=-1)
@click.option('-d', '--directory', flag_value=True, help='Only add directories, not files inside directories')
@click.option('-l', '--locale', help='If source locale is different from the default configuration. Use ltk list -l to see possible locales')
@click.option('-f', '--format',
              help="Format of file; if not specified, will use extension to detect; defaults to plaintext. Use ltk list -f to see possible formats. Files may not be added to Lingotek's system if not formatted correctly according to the specified format")
@click.option('-D', '--download_folder', type=click.Path(exists=True), help='Download folder for the translations for this file.  If set, it will take precedence over locale folders and the default download folder and will ignore clone as well as the no_ext argument when pulling/downloading')
@click.option('-o', '--overwrite', flag_value=True, help='Overwrite previously added file if the file has been modified')
@click.option('-s', '--srx', type=click.Path(exists=True), help='srx file')
@click.option('-si', '--srx_id', help='srx id')
@click.option('-i', '--its', type=click.Path(exists=True), help='its file')
@click.option('-ii', '--its_id', help='its id')
@click.option('-c', '--charset', help='File encoding')
@click.option('-ff', '--fprm', type=click.Path(exists=True), help='fprm file')
@click.option('-fi', '--fprm_id', help='fprm id')
@click.option('-fs', '--fprm_subfilter', type=click.Path(exists=True), help='fprm subfilter file')
@click.option('-fsi', '--fprm_subfilter_id', help='fprm subfilter id')
@click.option('-v', '--vault_id', help='Save-to TM vault id')
@click.option('-e', '--external_url', help='Source url')
@click.option('--due_date', help='Due date (as Unix timestamp, in milliseconds)')
@click.option('--due_reason', help='Reason for due date')
@click.option('-m', '--metadata', flag_value=True, help="Launches the metadata wizard")

# # Metadata - optional parameters
# @click.option('--author_email', help='Author email')
# @click.option('--author_name', help='Author name')
# @click.option('--business_division', help='Business division')
# @click.option('--business_unit', help='Business unit')
# @click.option('--campaign_id', help='Campaign ID')
# @click.option('--campaign_rating', help='Campaign rating')
# @click.option('--channel', help='Channel')
# @click.option('--contact_email', help='Contact email')
# @click.option('--contact_name', help='Contact name')
# @click.option('--content_description', help='Content description')
# @click.option('--content_type', help='Content type')
# @click.option('--domain', help='Domain')
# @click.option('--external_application_id', help='External application ID')
# @click.option('--external_document_id', help='External document ID')
# @click.option('--external_style_id', help='External style ID')
# @click.option('--job_id', help='Job ID')
# @click.option('--purchase_order', help='Purchase Order')
# @click.option('--reference_url', help='Reference URL')
# @click.option('--region', help='Region')
# @click.option('--require_review', help='Require review')
# @click.option('--category_id', help='Category ID')
# @click.option('--note', help='Note')

# Metadata - optional parameters
@click.option('--author_email', hidden=True)
@click.option('--author_name', hidden=True)
@click.option('--business_division', hidden=True)
@click.option('--business_unit', hidden=True)
@click.option('--campaign_id', hidden=True)
@click.option('--campaign_rating', hidden=True)
@click.option('--channel', hidden=True)
@click.option('--contact_email', hidden=True)
@click.option('--contact_name', hidden=True)
@click.option('--content_description', hidden=True)
@click.option('--content_type', hidden=True)
@click.option('--domain', hidden=True)
@click.option('--external_application_id', hidden=True)
@click.option('--external_document_id', hidden=True)
@click.option('--external_style_id', hidden=True)
@click.option('--job_id', hidden=True)
@click.option('--purchase_order', hidden=True)
@click.option('--reference_url', hidden=True)
@click.option('--region', hidden=True)
@click.option('--require_review', hidden=True)
@click.option('--category_id', hidden=True)
@click.option('--note', hidden=True)

def add(file_names, **kwargs):
    #""" Add files and folders for upload to Lingotek.  Fileglobs (e.g. *.txt) can be used to add all matching files and/or folders. Added folders will automatically add the new files added or created inside of them. """
    """ Add files and folders for upload to Lingotek.  Fileglobs (e.g. *.txt) can be used to add all matching files and/or folders. Added folders will automatically add the new files added or created inside of them.
    
    Metadata can be added by launching the metadata wizard with the -m flag or by using flags for specific metadata.  The metadata flags are --author_email, --author_name, --business_division, --business_unit, --campaign_id, --campaign_rating, --channel, --contact_email, --contact_name, --content_description, --content_type, --domain, --external_application_id, --external_document_id, --external_style_id, --job_id, --purchase_order, --reference_url, --region, --require_review, --category_id, and --note """
    try:
        action = add_action.AddAction(os.getcwd())
        init_logger(action.path)

        file_names = remove_powershell_formatting(file_names)

        for f in kwargs:
            if kwargs[f]:
                temp = remove_powershell_formatting(kwargs[f])
                kwargs[f] = temp

        action.add_action(file_names, **kwargs)
    except (UninitializedError, RequestFailedError, ResourceNotFound, AlreadyExistsError) as e:
        print_log(e)
        logger.error(e)
        return

@ltk.command(short_help="Sends updated content to Lingotek for documents that have been added; defaults to the entire project.")
@click.option('-n', '--test', 'test', flag_value=True, help='Shows which files will be added or updated without actually uploading any content')
@click.option('-t', '--title', 'title', flag_value=True, help='Display document titles rather than file paths')
@click.argument('files', type=click.Path(exists=True), required=False, nargs=-1)
@click.option('--due_date', help='Due date (as Unix timestamp, in milliseconds)')
@click.option('--due_reason', help='Reason for due date')
@click.option('-m', '--metadata', flag_value=True, help="Launches the metadata wizard")
@click.option('-o', '--metadata-only', 'metadata_only', flag_value=True, help="Only updates the metadata and due date/due reason and does not update the document contents")

# # Metadata - optional parameters
# @click.option('--author_email', help='Author email')
# @click.option('--author_name', help='Author name')
# @click.option('--business_division', help='Business division')
# @click.option('--business_unit', help='Business unit')
# @click.option('--campaign_id', help='Campaign ID')
# @click.option('--campaign_rating', help='Campaign rating')
# @click.option('--channel', help='Channel')
# @click.option('--contact_email', help='Contact email')
# @click.option('--contact_name', help='Contact name')
# @click.option('--content_description', help='Content description')
# @click.option('--content_type', help='Content type')
# @click.option('--domain', help='Domain')
# @click.option('--external_application_id', help='External application ID')
# @click.option('--external_document_id', help='External document ID')
# @click.option('--external_style_id', help='External style ID')
# @click.option('--job_id', help='Job ID')
# @click.option('--purchase_order', help='Purchase Order')
# @click.option('--reference_url', help='Reference URL')
# @click.option('--region', help='Region')
# @click.option('--require_review', help='Require review')
# @click.option('--category_id', help='Category ID')
# @click.option('--note', help='Note')

# Metadata - optional parameters
@click.option('--author_email', hidden=True)
@click.option('--author_name', hidden=True)
@click.option('--business_division', hidden=True)
@click.option('--business_unit', hidden=True)
@click.option('--campaign_id', hidden=True)
@click.option('--campaign_rating', hidden=True)
@click.option('--channel', hidden=True)
@click.option('--contact_email', hidden=True)
@click.option('--contact_name', hidden=True)
@click.option('--content_description', hidden=True)
@click.option('--content_type', hidden=True)
@click.option('--domain', hidden=True)
@click.option('--external_application_id', hidden=True)
@click.option('--external_document_id', hidden=True)
@click.option('--external_style_id', hidden=True)
@click.option('--job_id', hidden=True)
@click.option('--purchase_order', hidden=True)
@click.option('--reference_url', hidden=True)
@click.option('--region', hidden=True)
@click.option('--require_review', hidden=True)
@click.option('--category_id', hidden=True)
@click.option('--note', hidden=True)

def push(test, title, files, metadata, metadata_only, **kwargs):
    #""" Sends updated content to Lingotek for documents that have been added.  Fileglobs (e.g. *.txt) can be used to push all matching files """
    """ Sends updated content to Lingotek for documents that have been added.  Fileglobs (e.g. *.txt) can be used to push all matching files

    Metadata can be updated by launching the metadata wizard with the -m flag or by using flags for specific metadata.  The metadata flags are --author_email, --author_name, --business_division, --business_unit, --campaign_id, --campaign_rating, --channel, --contact_email, --contact_name, --content_description, --content_type, --domain, --external_application_id, --external_document_id, --external_style_id, --job_id, --purchase_order, --reference_url, --region, --require_review, --category_id, and --note """
    try:
        action = push_action.PushAction(os.getcwd(), test, title)
        init_logger(action.path)
        action.push_action(files=files, set_metadata=metadata, metadata_only=metadata_only, **kwargs)
    except UninitializedError as e:
        print_log(e)
        logger.error(e)
        return


@ltk.command(short_help="Add targets to document(s) to start translation; defaults to the entire project. Use ltk list -l to see possible locales")
@click.option('-n', '--doc_name', help='The name of the document for which to request target locale(s)')
@click.option('-p', '--path', type=click.Path(exists=True), help='The file name or directory for which to request target locale(s)')
@click.option('-c', '--cancel', 'to_cancel', flag_value=True, help='Cancels a specified target locale')
@click.option('-d', '--delete', 'to_delete', flag_value=True, help='Deletes a specified target locale')
@click.option('--due_date', help='The due date of the translation')
@click.option('-w', '--workflow', help='The workflow of the translation (Use "ltk list -w" to see available workflows)')
@click.argument('locales', required=False, nargs=-1)  # can have unlimited number of locales
def request(doc_name, path, locales, to_cancel, to_delete, due_date, workflow):
    """ Add targets to document(s) to start translation; defaults to the entire project. If no locales are specified, Filesystem Connector
        will look for target watch locales set in ltk config. Use ltk list -l to see possible locales. """
    try:
        action = request_action.RequestAction(os.getcwd(), doc_name, path, locales, to_cancel, to_delete, due_date, workflow)
        init_logger(action.path)
        if locales and isinstance(locales,str):
            locales = [locales]

        doc_name = remove_powershell_formatting(doc_name)
        path = remove_powershell_formatting(path)

        action.target_action()
    except (UninitializedError, ResourceNotFound, RequestFailedError) as e:
        print_log(e)
        logger.error(e)
        return

# todo add a --all option to see all document ids once only show relative to cwd is implemented
@ltk.command(name='list', short_help='Shows docs (default), workflows, locales, formats, or filters')
@click.option('-t', '--title', 'title', flag_value=True, help='List document titles and folder paths from project root instead of relative file paths')
@click.option('-c', '--hide_docs', 'hide_docs', flag_value=True, help='Collapse down to list only added directories instead of both directories and documents.')
@click.option('-w', '--workflows', 'id_type', flag_value='workflow', help='List available workflows')
@click.option('-l', '--locales', 'id_type', flag_value='locale', help='List supported locale codes')
@click.option('-f', '--formats', 'id_type', flag_value='format', help='List supported formats')
@click.option('-r', '--remote', 'id_type', flag_value='remote', help='List all project documents on Lingotek Cloud')
@click.option('--filters', 'id_type', flag_value='filter', help='List default and custom filters')
@click.option('-d', '--download_folder', 'show_dests', flag_value=True, help="Show target download folders for files that have had them set")
def list(**kwargs):
    """ Shows docs, workflows, locales, formats, or filters. By default lists added folders and docs. """
    try:
        action = list_action.ListAction(os.getcwd())
        init_logger(action.path)
        action.list_action(**kwargs)

    except (UninitializedError, RequestFailedError) as e:
        print_log(e)
        logger.error(e)
        return


@ltk.command(short_help="Gets the status of a specific document or all documents")
@click.option('-n', '--doc_name', help='Specific document name to get status of')
@click.option('-d', '--detailed', flag_value=True, help='Detailed status of each locale for the document')
@click.option('-a', '--all', flag_value=True, help='List all project documents on Lingotek Cloud')
def status(**kwargs):
    """ Gets the status of a specific document or all documents """
    try:
        action = status_action.StatusAction(os.getcwd())
        init_logger(action.path)

        for f in kwargs:
            if kwargs[f]:
                temp = remove_powershell_formatting(kwargs[f])
                kwargs[f] = temp

        action.get_status(**kwargs)
    except (UninitializedError, ResourceNotFound) as e:
        print_log(e)
        logger.error(e)
        return


@ltk.command(short_help='Download specified translations')
@click.option('-a', '--auto_format', flag_value=True, help='Flag to auto apply formatting during download')
@click.option('-l', '--locales', help="Specify locales to download (defaults to all target locales for the document). For multiple locales give a list separated by commas and no spaces (ex: en_US,en_GB).")
@click.option('-e', '--locale_ext', flag_value=True, help="Specifies to add the name of the locale as an extension to the file name (ex: doc1.fr_FR.docx). This is the default unless the clone download option is active.")
@click.option('-n', '--no_ext', flag_value=True, help="Specifies to not add the name of the locale as an extension to the file name. This is the default if the clone download option is active.")
@click.option('-x', '--xliff', flag_value=True, help="Download xliff version of the specified translation")
@click.argument('file_names', type=click.Path(exists=True), required=True, nargs=-1)
def download(auto_format, locales, locale_ext, no_ext, xliff, file_names):
    """ Downloads translated content specified by filename for specified locales, or all locales if none are specified. Change download options and folders using ltk config."""
    try:
        action = download_action.DownloadAction(os.getcwd())
        init_logger(action.path)

        for name in file_names:
            action.download_by_path(name, locales, locale_ext, no_ext, auto_format, xliff)
            print("\n")

    except (UninitializedError, ResourceNotFound, RequestFailedError) as e:
        print_log(e)
        logger.error(e)
        return


@ltk.command(short_help='Pulls translations for all added documents for all locales or by specified locales')
@click.option('-a', '--auto_format', flag_value=True, help='Flag to auto apply formatting during download')
@click.option('-e', '--locale_ext', flag_value=True, help="Specifies to add the name of the locale as an extension to the file name (ex: doc1.fr_FR.docx). This is the default unless the clone download option is active.")
@click.option('-n', '--no_ext', flag_value=True, help="Specifies to not add the name of the locale as an extension to the file name. This is the default if the clone download option is active.")
@click.argument('locales', nargs=-1)
def pull(auto_format, locale_ext, no_ext, locales):
    """ Pulls translations for all added documents for all locales or by specified locales """
    try:
        download = download_action.DownloadAction(os.getcwd())
        action = pull_action.PullAction(os.getcwd(), download)
        init_logger(action.path)
        if locales:
            for locale in locales:
                action.pull_translations(locale, locale_ext, no_ext, auto_format)
        else:
            action.pull_translations(None, locale_ext, no_ext, auto_format)
    except UninitializedError as e:
        print_log(e)
        logger.error(e)
        return


@ltk.command(name="rm", short_help="Disassociates local doc(s) from Lingotek Cloud and cancels the remote copy")
@click.argument('file_names', required=False, nargs=-1)
@click.option('-d', '--directory', flag_value=True, help='Only remove directories, not files inside directories')
@click.option('-i', '--id', flag_value=True, help='Cancels documents with the specified ids (instead of file names) on Lingotek Cloud.  Can be used to cancel documents that are not locally tracked.')
@click.option('-n', '--name', flag_value=True, help='Cancels documents with the specified names (instead of file names or paths) on Lingotek Cloud')
@click.option('-a', '--all', flag_value=True, help='Cancels all documents from Lingotek Cloud that are found locally')
@click.option('-l', '--local', flag_value=True, help='Deprecated.  Use \'ltk rm -f -a\' for all documents or \'ltk rm -f\' for specific documents.  (Legacy Usage: deletes all documents locally and cancels them in the Lingotek Cloud. Can be used in association with --name to delete a specified document locally.)')
@click.option('-r', '--remote', flag_value=True, help='Deletes specified documents from Lingotek Cloud instead of cancelling them')
@click.option('-f', '--force', flag_value=True, help='Delete local copy when cancelling remote source documents')
def rm(file_names, **kwargs):
    """
    Disassociates local doc(s) from Lingotek Cloud and removes them from the project by cancelling them.  If the remote copy should be deleted, use the -r flag.
    """
    try:
        action = rm_action.RmAction(os.getcwd())
        init_logger(action.path)
        if not file_names and not (('all' in kwargs and kwargs['all']) or ('local' in kwargs and kwargs['local'])):
            logger.info("Usage: ltk rm [OPTIONS] FILE_NAMES...")
            return

        if len(file_names) > 0:
            file_names = remove_powershell_formatting(file_names)

        action.rm_action(file_names, **kwargs)
    except (UninitializedError, ResourceNotFound, RequestFailedError) as e:
        print_log(e)
        logger.error(e)
        return

@ltk.command(short_help="Move file or directory at the specified location to a specified destination folder.")
@click.argument('source_path', type=click.Path(exists=True), required=True, nargs=-1)
@click.argument('destination_path', required=True, nargs=1)
def mv(source_path, destination_path):
    """
    Moves specified local doc to a specified destination directory, moving both the file itself and file location stored in the local database.
    If SOURCE_PATH is a directory, all added files in the directory will be moved.
    """
    try:
        # action = actions.Action(os.getcwd())
        add = add_action.AddAction(os.getcwd())
        action = move_action.MoveAction(add, os.getcwd())
        init_logger(action.path)

        source_path = remove_powershell_formatting(source_path)
        #print("Source path " + str(source_path))
        destination_path = remove_powershell_formatting(destination_path)
        #print("Destination path "+str(destination_path))

        action.mv_action(source_path, destination_path)
    except(UninitializedError, RequestFailedError) as e:
        print_log(e)
        logger.error(e)
        return

@ltk.command(name='import', short_help="Import docs from Lingotek")
@click.option('-a', '--all', 'import_all', flag_value=True, help='Import all documents from Lingotek Cloud')
@click.option('-f', '--force', flag_value=True, help='Overwrites existing documents without prompt')
@click.option('-p', '--path', type=click.Path(exists=True), help='Import documents to a specified path')
@click.option('-t', '--track', flag_value=True, help='Automatically add the imported documents to local tracking if they were not already being tracked.  Will not track cancelled documents.  Note: They will be added without extra options (no srx id, no download folder, etc.)')
@click.option('-c', '--no-cancel', flag_value=True, help='Don\'t include documents that have been cancelled.')
def import_command(import_all, force, path, track, no_cancel):
    """
    Import documents from Lingotek Cloud, by default downloading to the project's root folder
    """
    # todo import should show all documents
    # add a force option so can import all force -- overwrites all existing documents without prompting
    # check if doc id
    # if exist, prompt for overwrite
    # else automatically re-name
    # possibly have to patch title in Lingotek Cloud?
    try:
        # action = actions.Action(os.getcwd())
        action = import_action.ImportAction(os.getcwd())
        init_logger(action.path)

        if path != None:
            path = remove_powershell_formatting(path)

        action.import_action(import_all, force, path, track, no_cancel)
    except(UninitializedError, RequestFailedError) as e:
        print_log(e)
        logger.error(e)
        return


@ltk.command(short_help="Cleans up the associations between local documents and documents in Lingotek")
@click.option('-a', '--all', 'dis_all', flag_value=True, help='Removes all associations between local and remote documents and cancels those documents')
@click.argument('file_paths', required=False, nargs=-1)
@click.option('-f', '--force', flag_value=True, help='Deletes locally tracked documents that have been cancelled or no longer exist in Lingotek')
def clean(force, dis_all, file_paths):
    """
    Cleans up the associations between local documents and documents in Lingotek.
    By default, removes documents from local tracking that have been cancelled or no longer exist locally or in the Lingotek Cloud.
    Enter file or directory names to cancel those documents and remove local associations of specific files or directories.
    """
    try:
        action = clean_action.CleanAction(os.getcwd())
        init_logger(action.path)

        if len(file_paths) > 0:
            file_paths = remove_powershell_formatting(file_paths)

        action.clean_action(force, dis_all, file_paths)
    except (UninitializedError, RequestFailedError) as e:
        print_log(e)
        logger.error(e)
        return

@ltk.command(short_help="Copies added source folders for each locale")
@click.argument('folders', required=False, nargs=-1)
@click.option('-c', '--copy_root', flag_value=True, help='Copies the root source folder as a subfolder inside the locale folder, even if there is only one source folder being cloned.')
def clone(folders, copy_root):
    """
    Copies the folder structure of added folders or specified folders
    for each target locale as specified in config.
    Folders are added to the locale folder specified if one has been specified,
    or by default a new folder will be created with the name of the locale. If
    only one root folder is being cloned, then the locale folder is used
    (instead of creating a new folder inside of the locale folder).
    """
    try:
        action = clone_action.CloneAction(os.getcwd())
        init_logger(action.path)
        if isinstance(folders,str):
            folders = [folders]

        if len(folders) > 0:
            folders = remove_powershell_formatting(folders)

        action.clone_action(folders, copy_root)
    except (UninitializedError, RequestFailedError) as e:
        print_log(e)
        logger.error(e)
        return

@ltk.command(short_help="Watches local and remote files")
# @click.option('-p', '--path', type=click.Path(exists=True), multiple=True, help='Specify a folder to watch. Use option multiple times to specify multiple folders.')
@click.option('--ignore', multiple=True, help='Specify types of files to ignore.  For multiple types, specify this flag multiple times.  For example, to ignore pdf and html files, use "ltk watch --ignore .pdf --ignore .html"')
@click.option('--auto', 'delimiter', help='Automatically detects locale from the file name; specify locale delimiter')
@click.option('-t', '--timeout', type=click.INT, default=60,
              help='The amount of time watch will sleep between polls, in seconds. Defaults to 1 minute')
@click.option('-n','--no_folders', flag_value=True, help='Ignore files added to watch folders and only watch documents that have already been added.')
@click.option('-f','--force_poll', flag_value=True, help='Force API calls to Lingotek system at every poll for every document')
def watch(ignore, delimiter, timeout, no_folders, force_poll): # path, ignore, delimiter, timeout, no_folders):
    """
    Watches local files added by ltk, and sends a PATCH when a document is changed.
    Also watches remote files, and automatically downloads finished translations.
    Automatically adds documents that are added to the watchfolder.  Note: The add is performed without extra options (no srx id, no download folder, etc.)
    """
    try:
        action = WatchAction(os.getcwd(), timeout)
        init_logger(action.path)
        action.watch_action(ignore, delimiter, no_folders, force_poll) #path, ignore, delimiter, no_folders)
    except (UninitializedError, RequestFailedError) as e:
        print_log(e)
        logger.error(e)
        return

# Filters (Split into files, see http://bit.ly/2jArTRm)

@click.group(short_help="List, create, update, or delete Lingotek filters")
def filters():
    pass

@filters.command(name='add',short_help="Create a filter on Lingotek.")
@click.argument('filename')
@click.option('-t', '--type', 'filter_type', type=click.Choice(['FPRM','SRX','ITS']), help="The filter type being added.  Must be one of the following: FPRM, SRX, ITS.  When not explicitly specified, the file extension is used to attempt to detect the type.")
def filter_add(filename, filter_type):
    """Create filter on Lingotek."""
    try:
        action = filters_action.FiltersAction(os.getcwd())
        init_logger(action.path)
        action.filter_add_action(filename, filter_type)
    except (UninitializedError, RequestFailedError) as e:
        print_log(e)
        logger.error(e)
        return

@filters.command(name='save',short_help="Update filter on Lingotek.")
@click.argument('filter_id')
@click.argument('filename')
def filter_add(filter_id, filename):
    """Update filter on Lingotek."""
    try:
        action = filters_action.FiltersAction(os.getcwd())
        init_logger(action.path)
        action.filter_save_action(filter_id, filename)
    except (UninitializedError, RequestFailedError) as e:
        print_log(e)
        logger.error(e)
        return

@filters.command(name="get", short_help="Retrieve filter contents from Lingotek.")
@click.argument('filter_id')
@click.argument('filename', required=False)
@click.option('--info','info',flag_value=True, help="Retrieve filter info only.")
@click.option('--overwrite',flag_value=True, help="Overwrite local file when it already exists.")
def filter_get(filter_id, filename, info, overwrite):
    """Retrieve the filter specified by FILTER_ID from Lingotek and store it in the current working directly as the title (or as as the optional_filename when specified) of the filter"""
    try:
        action = filters_action.FiltersAction(os.getcwd())
        init_logger(action.path)
        if info == True:
            action.filter_info_action(filter_id)
        else:
            action.filter_get_action(filter_id, filename, overwrite)
    except (UninitializedError, RequestFailedError) as e:
        print_log(e)
        logger.error(e)
        return

@filters.command(name="list")
def filter_list():
    """List default and custom filters."""
    try:
        action = filters_action.FiltersAction(os.getcwd())
        init_logger(action.path)
        action.filter_list_action()
    except (UninitializedError, RequestFailedError) as e:
        print_log(e)
        logger.error(e)
        return

@filters.command(name="rm", short_help="Remove filter from Lingotek.")
@click.argument('filter_id')
def filter_rm(filter_id):
    """Remove the filter specified by FILTER_ID."""
    try:
        action = filters_action.FiltersAction(os.getcwd())
        init_logger(action.path)
        action.filter_rm_action(filter_id)
    except (UninitializedError, RequestFailedError) as e:
        print_log(e)
        logger.error(e)
        return

ltk.add_command(filters)

@click.group(short_help="Manages reference material attached to documents.")
def reference():
    pass

@reference.command(name='add', short_help="Uploads reference material and attaches it to the specified document.")
@click.argument('filename')
@click.option('-i', '--id', 'doc_id', flag_value=True, help="Adds reference material to the specified document ID instead of the specified filename.")
def reference_add(filename, doc_id):
    """Adds reference material to a document on Lingotek."""
    try:
        action = reference_action.ReferenceAction(os.getcwd())
        init_logger(action.path)
        action.reference_add_action(filename, doc_id)
    except (UninitializedError, RequestFailedError) as e:
        print_log(e)
        logger.error(e)
        return

@reference.command(name='list', short_help="Lists the reference material that is currently attached to the specified document.")
@click.argument('filename')
@click.option('-i', '--id', 'doc_id', flag_value=True, help="Lists reference material attached to the specified document ID instead of the specified filename.")
def reference_list(filename, doc_id):
    """Lists reference material attached to a document on Lingotek."""
    try:
        action = reference_action.ReferenceAction(os.getcwd())
        init_logger(action.path)
        action.reference_list_action(filename, doc_id)
    except (UninitializedError, RequestFailedError) as e:
        print_log(e)
        logger.error(e)
        return

@reference.command(name='get', short_help="Downloads reference material that is currently attached to the specified document.")
@click.argument('filename')
@click.option('-i', '--id', 'doc_id', flag_value=True, help="Downloads reference material attached to the specified document ID instead of the specified filename.")
@click.option('-a', '--all', 'get_all', flag_value=True, help="Skips the prompt and downloads all the reference material that is attached to the document.")
@click.option('-p', '--path', type=click.Path(exists=True), help='Download reference material to a specified path')
def reference_get(filename, doc_id, get_all, path):
    """Downloads reference material attached to a document on Lingotek.  Defaults to downloading them to the root of the project"""
    try:
        action = reference_action.ReferenceAction(os.getcwd())
        init_logger(action.path)
        action.reference_download_action(filename, doc_id, get_all, path)
    except (UninitializedError, RequestFailedError) as e:
        print_log(e)
        logger.error(e)
        return

@reference.command(name='rm', short_help="Removes reference material that is currently attached to the specified document")
@click.argument('filename')
@click.option('-i', '--id', 'doc_id', flag_value=True, help="Removes reference material from the specified document ID instead of the specified filename.")
@click.option('-a', '--all', 'remove_all', flag_value=True, help="Skips the prompt and removes all the reference material from the document.")
def reference_remove(filename, doc_id, remove_all):
    """Deletes reference material attached to a document on Lingotek."""
    try:
        action = reference_action.ReferenceAction(os.getcwd())
        init_logger(action.path)
        action.reference_remove_action(filename, doc_id, remove_all)
    except (UninitializedError, RequestFailedError) as e:
        print_log(e)
        logger.error(e)
        return

ltk.add_command(reference)

if __name__ == '__main__':
    ltk()



# coding: utf-8
from setuptools import setup, find_packages
import sys

if sys.version_info[0] == 2:
    base_dir = 'python2'
elif sys.version_info[0] == 3:
    base_dir = 'python3'

try:
    import pypandoc

    long_description = pypandoc.convert('README.md', 'rst')
    long_description = long_description.replace("\r", "")
except (IOError, ImportError, OSError):
    long_description = open('README.md').read()

setup(
    name='ltk',
    version='1.3.9',
    url='https://github.com/Lingotek/filesystem-connector',
    description='The Lingotek Filesystem Connector (ltk) links your files and folders to the Translation Network™',
    long_description=long_description,
    author='Lingotek',
    author_email='integrations@lingotek.com',
    license='MIT',
    packages=[ 'ltk',],
    package_dir={
        'ltk' : base_dir + '/ltk',
    },
    include_package_data=True,
    install_requires=[
        'click',
        'requests[security]',
        'tinydb',
        'watchdog',
        'gitpython',
        'pexpect',
        'blessings',
        'tabulate'
    ],
    extras_require={
        'security': ['pyOpenSSL', 'ndg-httpsclient', 'pyasn1'],
    },
    entry_points='''
        [console_scripts]
        ltk=ltk.commands:ltk
    ''',
)



# Using the following encoding: utf-8
import ctypes
from ltk.actions.action import Action
from ltk.actions import add_action
from ltk.actions import request_action
from ltk.actions import download_action
from ltk.logger import logger
from ltk.utils import map_locale, restart, get_relative_path, log_error
from ltk.locales import locale_list
import time
import requests
from requests.exceptions import ConnectionError
import os
import re
import sys
from watchdog.observers import Observer
from watchdog.events import FileSystemEvent
from ltk.watchhandler import WatchHandler
from ltk.git_auto import Git_Auto
import ltk.check_connection

DEFAULT_COMMIT_MESSAGE  = "Translations updated for "

# retry decorator to retry connections
def retry(logger, timeout=5, exec_type=None):
    if not exec_type:
        exec_type = [requests.exceptions.ConnectionError]

    def decorator(function):
        def wrapper(*args, **kwargs):
            while True:
                try:
                    return function(*args, **kwargs)
                except Exception as e:
                    log_error(self.error_file_name, e)
                    if e.__class__ in exec_type:
                        logger.error("Connection has timed out. Retrying..")
                        time.sleep(timeout)  # sleep for some time then retry
                    else:
                        raise e
        return wrapper
    return decorator

def has_hidden_attribute(file_path):
    """ Detects if a file has hidden attributes """
    try:
        # Python 2
        # attrs = ctypes.windll.kernel32.GetFileAttributesW(unicode(file_path))
        # End Python 2
        # Python 3
        attrs = ctypes.windll.kernel32.GetFileAttributesW(str(file_path))
        # End Python 3
        assert attrs != -1
        result = bool(attrs & 2)
    except (AttributeError, AssertionError):
        result = False
    return result

class WatchAction(Action):
    # def __init__(self, path, remote=False):
    def __init__(self, path=None, timeout=60):
        Action.__init__(self, path, True, timeout)
        self.observers = []  # watchdog observers that will watch the files
        self.handler = WatchHandler()
        self.handler.on_modified = self._on_modified
        self.handler.on_created = self._on_created
        self.handler.on_moved = self._on_moved
        self.watch_queue = []  # not much slower than deque unless expecting 100+ items
        self.locale_delimiter = None
        self.ignore_ext = []  # file types to ignore as specified by the user
        self.detected_locales = {}  # dict to keep track of detected locales
        self.watch_folder = True
        self.timeout = timeout
        self.updated = {}
        self.git_auto = Git_Auto(path)
        self.polled_list = set([])
        self.force_poll = False
        self.add = add_action.AddAction(path)
        self.download = download_action.DownloadAction(path)
        self.root_path = path
        self.download_file_paths = set() # set of download file path names
        # if remote:  # poll lingotek cloud periodically if this option enabled
        # self.remote_thread = threading.Thread(target=self.poll_remote(), args=())
        # self.remote_thread.daemon = True
        # self.remote_thread.start()

    def is_hidden_file(self, file_path):
        # todo more robust checking for OSX files that doesn't start with '.'
        name = os.path.abspath(file_path).replace(self.path, "")
        if has_hidden_attribute(file_path) or ('Thumbs.db' in file_path) or ('ehthumbs.db' in file_path) or ('desktop.ini' in file_path):
            return True
        while name != "":
            if name.startswith('.') or name.startswith('~') or name == "4913":
                return True
            name = name.split(os.sep)[1:]
            name = (os.sep).join(name)
        return False

    def is_translation(self, file_name):
        locales = locale_list
        if any('.'+locale in file_name for locale in locales):
            locales = {v:k for k,v in enumerate(locales) if v in file_name}.keys()
            replace_target = None
            for locale in locales:
                original = file_name
                file_name = file_name.replace('.'+locale, '')
                if file_name != original:
                    replace_target = locale
                    break
            file_name = re.sub('\.{2,}', '.', file_name)
            file_name = file_name.rstrip('.')
            doc = self.doc_manager.get_doc_by_prop('file_name', file_name.replace(self.path, ''))
            replace_target = replace_target.replace("-", "_")
            if doc:
                if 'locales' in doc and replace_target in doc['locales']:
                    return True
        return False

    def check_remote_doc_exist(self, fn, document_id=None):
        """ check if a document exists remotely """
        if not document_id:
            entry = self.doc_manager.get_doc_by_prop('file_name', fn)
            document_id = entry['id']
        response = self.api.get_document(document_id)
        if response.status_code != 200:
            return False
        return True

    def _on_modified(self, event):
        """ Notify Lingotek cloud when a previously added file is modified """
        try:
            db_entries = self.doc_manager.get_all_entries()
            in_db = False
            fn = ''
            # print event.src_path
            for entry in db_entries:
                # print entry['file_name']
                if event.src_path.endswith(entry['file_name']):
                    fn = entry['file_name']
                    in_db = True
            if not event.is_directory and in_db:
                try:
                    # check that document is added in TMS before updating
                    if self.check_remote_doc_exist(fn) and self.doc_manager.is_doc_modified(fn, self.path):
                        #logger.info('Detected local content modified: {0}'.format(fn))
                        #self.update_document_action(os.path.join(self.path, fn))
                        #logger.info('Updating remote content: {0}'.format(fn))
                        try:
                            self.polled_list.remove(fn)
                        except Exception:
                            pass
                        self.update_content(fn)
                except KeyboardInterrupt:
                    for observer in self.observers:
                        observer.stop()
                except ConnectionError:
                    print("Could not connect to remote server.")
                    restart()
                except ValueError:
                    print(sys.exc_info()[1])
                    restart()
        except KeyboardInterrupt:
            for observer in self.observers:
                observer.stop()
        except Exception as err:
            restart("Error on modified: "+str(err)+"\nRestarting watch.")

    def _on_created(self, event):
        # get path
        # add action
        try:
            db_entries = self.doc_manager.get_all_entries()
            in_db = False
            fn = ''
            for entry in db_entries:
                if event.src_path.endswith(entry['file_name']):
                    fn = entry['file_name']
                    in_db = True
            if not event.is_directory and in_db:
                self._on_modified(event)
            else:
                file_path = event.src_path
                # if created file was a downloaded translation, don't add to poll. Prevents recursion from downloaded translation files when ltk watch is running
                if file_path in self.download_file_paths:
                    self.download_file_paths.remove(file_path)
                    return
                # if it's a hidden document, don't do anything
                if not self.is_hidden_file(file_path) and not self.is_translation(file_path):
                    relative_path = file_path.replace(self.path, '')
                    title = os.path.basename(os.path.normpath(file_path))
                    curr_ext = os.path.splitext(file_path)[1]
                    # return if the extension should be ignored or if the path is not a file
                    if curr_ext in self.ignore_ext or not os.path.isfile(file_path):
                        # logger.info("Detected a file with an extension in the ignore list, ignoring..")
                        return
                    # only add or update the document if it's not a hidden document and it's a new file
                    try:
                        if self.doc_manager.is_doc_new(relative_path, self.root_path) and self.watch_folder:
                            #testing
                            #self.polled_list.add(relative_path) #test that this doesn't break other areas of watch
                            #end testing

                            self.add_document(file_path, title, locale=self.locale)

                        elif self.doc_manager.is_doc_modified(relative_path, self.path):
                            self.update_content(relative_path)
                        else:
                            return
                    except KeyboardInterrupt:
                        for observer in self.observers:
                            observer.stop()
                    except ConnectionError:
                        print("Could not connect to remote server.")
                        restart()
                    except ValueError:
                        print(sys.exc_info()[1])
                        restart()
                    doc = self.doc_manager.get_doc_by_prop('file_name', relative_path)
                    if doc:
                        document_id = doc['id']
                    else:
                        return
                    if self.locale_delimiter:
                        try:
                            # curr_locale = title.split(self.locale_delimiter)[1]
                            # todo locale detection needs to be more robust
                            curr_locale = title.split(self.locale_delimiter)[-2]
                            fixed_locale = map_locale(curr_locale)
                            if fixed_locale:
                                print ("fixed locale: ", fixed_locale)
                                # self.watch_locales.add(fixed_locale)
                                self.detected_locales[document_id] = fixed_locale
                            else:
                                logger.warning('This document\'s detected locale: {0} is not supported.'.format(curr_locale))
                        except IndexError:
                            logger.warning('Cannot detect locales from file: {0}, not adding any locales'.format(title))
                    self.watch_add_target(relative_path, document_id)
                    # logger.info('Added new document {0}'.format(title
                # else:
                #     print("Skipping hidden file "+file_path)
        except KeyboardInterrupt:
            for observer in self.observers:
                observer.stop()
        # except Exception as err:
        #     restart("Error on created: "+str(err)+"\nRestarting watch.")

    def _on_moved(self, event):
        """Used for programs, such as gedit, that modify documents by moving (overwriting)
        the previous document with the temporary file. Only the moved event contains the name of the
        destination file."""
        try:
            event = FileSystemEvent(event.dest_path)
            self._on_modified(event)
        except KeyboardInterrupt:
            for observer in self.observers:
                observer.stop()
        except Exception as err:
            restart("Error on moved: "+str(err)+"\nRestarting watch.")

    def get_watch_locales(self, document_id):
        """ determine the locales that should be added for a watched doc """
        locales = []
        if self.detected_locales:
            try:
                locales = [self.detected_locales[document_id]]
            except KeyError:
                logger.error("Something went wrong. Could not detect a locale")
            return locales
        entry = self.doc_manager.get_doc_by_prop("id", document_id)
        try:
            locales = [locale for locale in self.watch_locales if locale not in entry['locales']]
        except KeyError:
            locales = self.watch_locales
        return locales

    def watch_add_target(self, file_name, document_id):
        if not file_name:
            title=self.doc_manager.get_doc_by_prop("id", document_id)
        else:
            title = os.path.basename(file_name)
        if document_id not in self.watch_queue:
            self.watch_queue.append(document_id)
        # Only add target if doc exists on the cloud
        if self.check_remote_doc_exist(title, document_id):
            locales_to_add = self.get_watch_locales(document_id)
            if locales_to_add == ['[]']: locales_to_add = []
            # if len(locales_to_add) == 1:
            #     printStr = "Adding target "+locales_to_add[0]
            # else:
            #     printStr = "Adding targets "
            #     for target in locales_to_add:
            #         printStr += target+","
            # print(printStr)
            if self.api.get_document(document_id):
                request = request_action.RequestAction(self.path, title, file_name, locales_to_add, None, None, None, None, document_id, True)
                if request.target_action() and document_id in self.watch_queue:
                    self.watch_queue.remove(document_id)

    def process_queue(self):
        """do stuff with documents in queue (currently just add targets)"""
        # todo may want to process more than 1 item every "poll"..
        # if self.watch_queue:
        #     self.watch_add_target(None, self.watch_queue.pop(0))
        for document_id in self.watch_queue:
            self.watch_add_target(None, document_id)

    def update_content(self, relative_path):
        if self.update_document_action(os.path.join(self.path, relative_path)):
            self.updated[relative_path] = 0
            logger.info('Updating remote content: {0}'.format(relative_path))

    def check_modified(self, doc): # Checks if the version of a document on Lingotek's system is more recent than the local version
        old_date = doc['last_mod']
        response = self.api.get_document(doc['id'])
        if response.status_code == 200:
            new_date = response.json()['properties']['modified_date']
            # orig_count = response.json()['entities'][1]['properties']['count']['character']
        else:
            print("Document not found on Lingotek Cloud: "+str(doc['name']))
            return False
        if int(old_date)<int(str(new_date)[0:10]):
            return True
        return False

    @retry(logger)
    def poll_remote(self):
        """ poll lingotek servers to check if translation is finished """
        if self.auto_format_option == 'on':
            autoFormat = True;
        else:
            autoFormat = False;
        documents = self.doc_manager.get_all_entries()  # todo this gets all documents, not necessarily only ones in watch folder
        # documents_downloaded = False
        git_commit_message = DEFAULT_COMMIT_MESSAGE
        for doc in documents:
            doc_id = doc['id']
            if doc_id in self.watch_queue:
                # if doc id in queue, not imported yet
                continue
            file_name = doc['file_name']
            # Wait for Lingotek's system to no longer show translation as as completed
            if file_name in self.updated:
                if self.updated[file_name] > 3:
                    self.updated.pop(file_name, None)
                else:
                    self.updated[file_name] += self.timeout
                    continue
            try:
                downloaded = doc['downloaded']
            except KeyError:
                downloaded = []
                self.doc_manager.update_document('downloaded', downloaded, doc_id)
            if file_name not in self.polled_list or self.force_poll:
                locale_progress = self.import_locale_info(doc_id, poll=True)
                # Python 2
                # for locale, progress in locale_progress.iteritems():
                # End Python 2
                # Python 3
                for locale in locale_progress:
                    progress = locale_progress[locale]
                # End Python 3
                    if progress == 100 and locale not in downloaded:
                        # document_added = False
                        # if (doc['name']+": ") not in git_commit_message:
                        #     if documents_downloaded: git_commit_message += '; '
                        #     git_commit_message += doc['name'] + ": "
                        #     document_added = True
                        # if document_added:
                        #     git_commit_message += locale
                        # else:
                        #     git_commit_message += ', ' + locale
                        # documents_downloaded = True
                        logger.info('Translation completed ({0} - {1})\n'.format(file_name, locale))
                        if self.locale_delimiter:
                            locale = locale.replace('_','-')
                            download_file_path = self.download.download_action(doc_id, locale, autoFormat, xliff=False, locale_ext=False)
                        else:
                            locale = locale.replace('_','-')
                            if self.clone_option == 'on':
                                # This prevents recursion when clone option is turned on and a file in a subfolder is added/modified while watch is running
                                download_file_path = self.download.download_action(doc_id, locale, autoFormat, xliff=False, locale_ext=False)
                                # Making clone paths a set prevents the list of paths to become full of duplicates from files added with the same name.
                                # Note that in the case of a file with the same name, this is only called if the file was created outside of the watch home dir first (ex. creating testing/test.txt then test.txt will trigger, creating test.txt then testing/test.txt will not)
                            else:
                                download_file_path = self.download.download_action(doc_id, locale, autoFormat)
                        # This prevents recursion when clone option is off and download folder is set. 
                        self.download_file_paths.add(download_file_path)
                    elif progress != 100 and locale in downloaded:
                        # print("Locale "+str(locale)+" for document "+doc['name']+" is no longer completed.")
                        self.doc_manager.remove_element_in_prop(doc_id, 'downloaded', locale)
                if set(locale_progress.keys()) == set(downloaded):
                    if all(value == 100 for value in locale_progress.values()):
                        self.polled_list.add(file_name)
        ###
        ### Adding, committing, and pushing to Github has been moved to DownloadAction ###
        ###
        # config_file_name, conf_parser = self.init_config_file()
        # git_autocommit = conf_parser.get('main', 'git_autocommit')
        # if git_autocommit in ['True', 'on'] and documents_downloaded == True:
        #     if self.git_auto.repo_exists():
        #         self.git_auto.commit(git_commit_message)
        #         self.git_auto.push()



    def complete_path(self, file_location):
        # print("self.path: "+self.path)
        # print("file_location: "+file_location)
        # print("abs file_location: "+os.path.abspath(file_location))
        abspath=os.path.abspath(file_location)
        # norm_path = os.path.abspath(file_location).replace(self.path, '')
        print
        return abspath.rstrip(os.sep)
       
    def watch_action(self, ignore, delimiter=None, no_folders=False, force_poll=False): # watch_paths, ignore, delimiter=None, no_folders=False):
        # print self.path
        watch_paths = None
        if not watch_paths:
            watch_paths = self.folder_manager.get_file_names()
            for i in range(len(watch_paths)):
                watch_paths[i] = get_relative_path(self.path, watch_paths[i])
        else:
            watch_paths_list = []
            for path in watch_paths:
                watch_paths_list.append(path.rstrip(os.sep))
            watch_paths = watch_paths_list
        if len(watch_paths) and not no_folders:
            self.watch_folder = True
        else:
            watch_paths = [os.getcwd()]
        if self.watch_folder:
            watch_message = "Watching for updates in "
            for i in range(len(watch_paths)):
                watch_paths[i] = self.complete_path(watch_paths[i])
                watch_message += "{0}".format(watch_paths[i])
                if i < len(watch_paths)-1:
                    watch_message += " "
            print (watch_message)
        else:
            print ("Watching for updates to added documents")
        if force_poll:
            self.force_poll = True
        self.ignore_ext.extend(ignore)
        self.locale_delimiter = delimiter
        for watch_path in watch_paths:
            observer = Observer()
            observer.schedule(self.handler, path=watch_path, recursive=True)
            try:
                observer.start()
            except OSError as e:
                logger.warning("Watching too many items, please be more specific by using ltk add on the files and folders that should be watched")
                return
            self.observers.append(observer)
        queue_timeout = 3
        # start_time = time.clock()
        try:
            while True:
                if ltk.check_connection.check_for_connection():
                    self.poll_remote()
                    current_timeout = self.timeout
                    while len(self.watch_queue) and current_timeout > 0:
                        self.process_queue()
                        time.sleep(queue_timeout)
                        current_timeout -= queue_timeout
                time.sleep(current_timeout) # default 60 sec total, 3 already taken
        except KeyboardInterrupt:
            for observer in self.observers:
                observer.stop()
        # except Exception as err:
        #     restart("Error: "+str(err)+"\nRestarting watch.")
        for observer in self.observers:
            observer.join()



from ltk.actions.action import *

class RequestAction(Action):
    def __init__(self, actionPath, doc_name, requestPath, entered_locales, to_cancel, to_delete, due_date, workflow, document_id=None, surpressMessage=False):
        Action.__init__(self, actionPath)

        self.document_name = doc_name
        self.requestPath = requestPath
        self.entered_locales = entered_locales
        self.to_cancel = to_cancel
        self.to_delete = to_delete
        self.workflow = workflow
        self.due_date = due_date
        self.document_id = document_id
        self.surpressMessage = surpressMessage

        self.expected_code = ''
        self.failure_message = ''
        self.info_message = ''
        self.change_db_entry = ''
        self.docs = []

    def target_action(self):
        try:
            is_successful = False
            locales = []
            if self.entered_locales:
                for locale in self.entered_locales:
                    locales.extend(locale.split(','))
                if self.to_delete:
                    locales = get_valid_locales(self.api, locales, 'deleted')
                elif self.to_cancel:
                    locales = get_valid_locales(self.api, locales, 'cancelled')
                else:
                    locales = get_valid_locales(self.api, locales, 'added')
            elif self.watch_locales:
                locales = self.watch_locales
            elif self.surpressMessage:
                # don't print out anything when on watch
                return False
            else:
                logger.info('No locales have been set. Locales can be passed in as arguments or set as target locales in ltk config.')
                return False
            if self.requestPath:
                self.document_id = None
                self.document_name = None
            self.change_db_entry = True
            if self.to_delete:
                if not self.entered_locales:
                    logger.error("Please enter a target locale to delete")
                    return
                self.expected_code = 204
                self.failure_message = 'Failed to delete target'
                self.info_message = 'Deleted locale'
            elif self.to_cancel:
                if not self.entered_locales:
                    logger.error("Please enter a target locale to cancel")
                    return
                self.expected_code = 204
                self.failure_message = 'Failed to cancel target'
                self.info_message = 'Cancelled locale'
            else:
                self.expected_code = 201
                self.failure_message = 'Failed to add target'
                self.info_message = 'Added target'
            # print("doc_name: "+str(document_name))
            if not self.requestPath and not self.document_name and not self.document_id:
                self.docs = self.doc_manager.get_all_entries()
            elif self.requestPath:
                self.docs = self.get_docs_in_path(self.requestPath)
            else:
                if self.document_id:
                    entry = self.doc_manager.get_doc_by_prop('id', self.document_id)
                    if not entry:
                        logger.error('Document specified for target doesn\'t exist: {0}'.format(self.document_id))
                        return
                elif self.document_name:
                    entry = self.doc_manager.get_doc_by_prop('name', self.document_name)
                    if not entry:
                        logger.error('Document name specified for target doesn\'t exist: {0}'.format(self.document_name))
                        return
                        # raise exceptions.ResourceNotFound("Document name specified doesn't exist: {0}".format(document_name))
                if not entry:
                    logger.error('Could not add target. File specified is invalid.')
                    return
                self.docs.append(entry)
            if len(self.docs) == 0:
                if self.requestPath and len(self.requestPath) > 0:
                    logger.info("File "+str(self.requestPath)+" not found")
                else:
                    logger.info("No documents to request a target locale")

            is_successful = self._request_translations(locales)
            return is_successful
        except Exception as e:
            log_error(self.error_file_name, e)
            if 'string indices must be integers' in str(e) or 'Expecting value: line 1 column 1' in str(e):
                logger.error("Error connecting to Lingotek's TMS")
            else:
                logger.error("Error on request: "+str(e))

    def _request_translations(self, locales):
        for entry in self.docs:
            if not self._handleEntryRequest(entry, locales):
                return False
        return True
       

    def _handleEntryRequest(self, entry, locales):
        is_successful = False
        self.document_id = entry['id']
        self.document_name = entry['file_name']
        existing_locales = []
        if 'locales' in entry and entry['locales']:
            existing_locales = entry['locales']
        for locale in locales:
            locale = locale.replace('_','-')
            if self.to_delete:
                response = self.api.document_delete_target(self.document_id, locale)
            elif self.to_cancel:
                response = self.api.document_cancel_target(self.document_id, locale)
            else:
                response = self.api.document_add_target(self.document_id, locale, self.workflow, self.due_date)
            if self.expected_code == 201: # Request target
                if response.status_code == 202:
                    if 'next_document_id' in response.json():
                        self.doc_manager.update_document('id', response.json()['next_document_id'], self.document_id)
                        self.document_id = response.json()['next_document_id']
                elif response.status_code == 402:
                    raise_error(response.json(), "", True)
                elif response.status_code == 410:
                    logger.info("Document has been archived. Reuploading...")
                    target_locales = existing_locales
                    if locale not in existing_locales:
                        target_locales.append(locale)
                    title = entry['name']
                    self.doc_manager.remove_element(self.document_id)
                    self.add_document(self.document_name, title, self.default_metadata, translation_locale_code=target_locales)
                    return
                elif response.status_code == 423:
                    if 'next_document_id' in response.json():
                        self.doc_manager.update_document('id', response.json()['next_document_id'], self.document_id)
                        self.document_id = response.json()['next_document_id']
                        self._handleEntryRequest(entry, [locale])
                    else:
                        raise_error(response.json(), "Document was locked but no next document id provided", False)
            if response.status_code != self.expected_code:
                if (response.json() and response.json()['messages']):
                    response_message = response.json()['messages'][0]
                    response_message = response_message.replace(self.document_id, self.document_name + ' (' + self.document_id + ')')
                    response_message = response_message.replace('.', ' ')
                    response_message = response_message + 'for document ' + self.document_name
                    print(response_message + "\n")
                    #if 'not found' in response_message:
                        #return
                else:
                    raise_error(response.json(), '{message} {locale} for document {name}\n'.format(message=self.failure_message, locale=locale, name=self.document_name), True)
                if not 'already exists' in response_message:
                    self.change_db_entry = False
                # self.update_doc_locales(document_id)
                continue
            logger.info('{message} {locale} for document {name}\n'.format(message=self.info_message, locale=locale, name=self.document_name))
        remote_locales = self.get_doc_locales(self.document_id, self.document_name) # Get locales from Lingotek Cloud
        locales_to_add = []
        existing_locales = []
        if 'locales' in entry and entry['locales']:
            existing_locales = entry['locales']
        if self.change_db_entry:
            # Make sure that the locales that were just added are added to the database as well as the previous remote locales (since they were only just recently added to Lingotek's system)
            if (self.to_delete or self.to_cancel) and self.entered_locales:
                locales_to_add = locales
            else:
                if remote_locales:
                    for locale in remote_locales:
                        if locale not in locales:
                            locales_to_add.append(locale)

                for locale in locales:
                    locale = locale.replace('_', '-')
                    if locale not in existing_locales and locale not in locales_to_add:
                        locales_to_add.append(locale)
            if self.to_delete:
                self._target_action_db(self.to_delete, locales_to_add, self.document_id)
            else:
                self._target_action_db(self.to_cancel, locales_to_add, self.document_id)
            #self._target_action_db((self.to_delete or self.to_cancel), locales_to_add, self.document_id)
            is_successful = True

        return is_successful


from ltk.actions.action import *

class PushAction(Action):
    def __init__(self, path, test, title):
        Action.__init__(self, path)
        self.title = title
        self.test = test

    def push_action(self, files=None, set_metadata=False, metadata_only=False, **kwargs):
        self.metadata_only = metadata_only
        self.metadata = copy.deepcopy(self.default_metadata)
        if set_metadata:
            self.metadata = self.metadata_wizard()
        elif self.metadata_prompt:
            if yes_no_prompt('Would you like to launch the metadata wizard?', default_yes=True):
                self.metadata = self.metadata_wizard()
        try:
            if files:
                added, updated, failed = self._push_specific_files(files, **kwargs)
            else:
                added = self._add_new_docs(**kwargs)
                updated, failed = self._update_current_docs(**kwargs)
            total = added + updated + failed
            if total is 0:
                report = 'All documents up-to-date with Lingotek Cloud. '
            else:
                report = "Added {0}, Updated {1}, Failed {2} (Total {3})".format(added, updated, failed, total)
            if self.test:
                logger.info("TEST RUN: " + report)
            else:
                logger.info(report)

        except Exception as e:
            log_error(self.error_file_name, e)
            if 'string indices must be integers' in str(e) or 'Expecting value: line 1 column 1' in str(e):
                logger.error("Error connecting to Lingotek's TMS")
            else:
                logger.error("Error on push: "+str(e))

    def _add_new_docs(self, **kwargs):
        folders = self.folder_manager.get_file_names()
        added = 0
        if len(folders) and not self.metadata_only:
            for folder in folders:
                matched_files = get_files(folder)
                if matched_files:
                    for file_name in matched_files:
                        try:
                            relative_path = self.norm_path(file_name)
                            title = os.path.basename(relative_path)
                            if self.doc_manager.is_doc_new(relative_path) and not self.doc_manager.is_translation(relative_path, title, matched_files, self):
                                display_name = title if self.title else relative_path
                                if self.test:
                                    print('Add {0}'.format(display_name))
                                else:
                                    self.add_document(file_name, title, doc_metadata=self.metadata, **kwargs)
                                added += 1
                        except json.decoder.JSONDecodeError as e:
                            log_error(self.error_file_name, e)
                            logger.error("JSON error on adding document.")
        return added

    def _update_current_docs(self, **kwargs):
        updated = 0
        failed = 0
        entries = self.doc_manager.get_all_entries()
        for entry in entries:
            if (len(self.metadata) > 0 or kwargs['due_date'] or kwargs['due_reason']) or (self.doc_manager.is_doc_modified(entry['file_name'], self.path) and not self.metadata_only):
                display_name = entry['name'] if self.title else entry['file_name']
                updated, failed = self._handle_update(updated, failed, display_name, entry, **kwargs)
        return updated, failed

    def _push_specific_files(self, patterns, **kwargs):
        files = set()
        added = 0
        updated = 0
        failed = 0
        for pattern in patterns:
            if os.path.isdir(pattern):
                for file in get_files(pattern):
                    relative_path = self.norm_path(file)
                    files.add(relative_path)
            else:
                relative_path = self.norm_path(pattern)
                files.add(relative_path)
        for file in files:
            title = os.path.basename(file)
            if not self.metadata_only and self.doc_manager.is_doc_new(file) and not self.doc_manager.is_translation(file, title, files, self):
                display_name = title if self.title else file
                if self.test:
                    print('Add {0}'.format(display_name))
                else:
                    self.add_document(file, title, doc_metadata=self.metadata, **kwargs)
                added += 1
            elif (len(self.metadata) > 0 or kwargs['due_date'] or kwargs['due_reason']) or (self.doc_manager.is_doc_modified(file, self.path) and not self.metadata_only):
                entry = self.doc_manager.get_doc_by_prop('file_name', file)
                if entry:
                    display_name = entry['name'] if self.title else entry['file_name']
                    updated, failed = self._handle_update(updated, failed, display_name, entry, **kwargs)
        return added, updated, failed

    def _handle_update(self, updated, failed, display_name, entry, **kwargs):
        if self.test:
            updated += 1 # would be updated
            print('Update {0}'.format(display_name))
            return updated, failed
        if self.update_document_action(entry['file_name'], display_name, doc_metadata=self.metadata, **kwargs):
            updated += 1
            logger.info('Updated {0}'.format(display_name))
        else:
            failed += 1
        return updated, failed


import os, sys
from ltk.locales import default_locales, locale_list
from ltk.logger import logger
import time
import logging
import traceback
from ltk import exceptions
try:
    from blessings import Terminal
    term = Terminal()
except ImportError:
    term = False
# from constants import APP_ID

class Enum(set):
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

# def get_access_token(host, username, password):
#     auth_uri = host + '/auth/authorize.html'
#     auth_params = {'client_id': APP_ID, 'redirect_uri': '', 'response_type': 'token'}

# todo possibly put dictionary outside so not built with every function call
def detect_format(file_name, get_mapper=False):
    format_mapper = {
        '.csv': 'CSV',
        '.dita': 'DITA',
        '.ditamap': 'DITAMAP',
        '.docx': 'DOCX_OKAPI',
        '.doc' : 'WORD_OKAPI',
        '.dtd': 'DTD',
        '.xlsx': 'XLSX_OKAPI',
        '.idml': 'IDML',
        '.properties': 'JAVA_PROPERTIES_OKAPI',
        '.json': 'JSON',
        '.pdf': 'PDF',
        '.txt': 'PLAINTEXT_OKAPI',
        '.po': 'PO',
        '.ppt': 'PPT_OKAPI',
        '.pptx': 'PPTX_OKAPI',
        '.resx': 'RESX',
        '.regex': 'REGEX',
        '.rtf': 'RTF_OKAPI',
        '.srt': 'SUBTITLE_RIP',
        '.strings': 'APPLE_STRINGS',
        '.tsv': 'TABLE',  # catkeys?
        '.ts': 'TS',
        '.xml': 'XML_OKAPI'
    }

    format_mapper.update(dict.fromkeys(['.dox', '.c', '.h', '.cpp'], 'DOXYGEN'))
    format_mapper.update(dict.fromkeys(['.html', '.htm'], 'HTML_OKAPI'))
    format_mapper.update(dict.fromkeys(['.odp', '.otp'], 'ODP'))
    format_mapper.update(dict.fromkeys(['.ods', '.ots'], 'ODS'))
    format_mapper.update(dict.fromkeys(['.odt', '.ott'], 'ODT'))
    format_mapper.update(dict.fromkeys(['.yaml', '.yml'], 'RAILS_YAML'))
    format_mapper.update(dict.fromkeys(['.xliff', '.xlf'], 'XLIFF_OKAPI'))
    if get_mapper:
        return format_mapper
    name, extension = os.path.splitext(file_name)
    return format_mapper.get(extension.lower(), 'PLAINTEXT_OKAPI')

def map_locale(locale):
    """
    maps incorrectly formatted locales to valid locales for use with Lingotek API
    :param locale: incorrectly formatted locale
    :return: valid locale
    """
    # import json
    # valid_locales = []
    # unsupported_locales = []
    # with open('data/language-default-locales.json') as json_file:
    #     default_locales = json.load(json_file)
    try:
        return default_locales[locale]
    except KeyError:
        return None
    # for locale in locales:
    #     try:
    #         valid_locales.append(default_locales[locale])
    #     except KeyError:
    #         unsupported_locales.append(locale)
    # return valid_locales, unsupported_locales

def restart(message="Restarting watch", interval=5):
    """Restarts the program. Used after exceptions. Otherwise, watch doesn't work anymore."""
    time.sleep(interval)
    print(message)
    cmd = "ltk"
    for arg in sys.argv[1:]:
        cmd = cmd + " " + arg
    os.system(cmd)

    ''' This way (below) works for Linux, but does not work on Windows '''
    #python = sys.executable
    #os.execl(python, python, * sys.argv)

def is_valid_locale(api, locale):
    """Returns true if the locale is found in Lingotek's remote list of locales or, if the api call fails, if the locale is found in the local list of locales."""
    valid_locales = []
    response = api.list_locales()
    remote_check = False
    if response.status_code == 200:
        remote_check = True
    locale_json = response.json()
    for entry in locale_json:
        valid_locales.append(locale_json[entry]['locale'])
    locales = []
    check_locale = locale.replace("-","_")
    if remote_check and check_locale not in valid_locales or not remote_check and not check_locale in locale_list:
        return False
    else:
        return True

def get_valid_locales(api, entered_locales, operation_text):
    """Return the list of valid locales, checking locales either remotely or using a local list of locales."""
    valid_locales = []
    response = api.list_locales()
    remote_check = False
    if response.status_code == 200:
        remote_check = True
    locale_json = response.json()
    for entry in locale_json:
        valid_locales.append(locale_json[entry]['locale'])
    locales = []
    if(len(entered_locales) == 0 or (len(entered_locales) == 1 and entered_locales[0] == "[]")):
        logger.warning('No locales have been assigned to this document.  Please add them using \'ltk request\'.')
    else:
        for locale in entered_locales:
            check_locale = locale.replace("-","_")
            if remote_check and check_locale not in valid_locales or not remote_check and not check_locale in locale_list:
                logger.warning('The locale code "'+str(locale)+'" failed to be '+operation_text+' since it is invalid (see "ltk list -l" for the list of valid codes).')
            else:
                locales.append(locale)
    return locales

def get_translation_files(file_name, path, download_option, doc_manager):
    translation_files = []

    if download_option == "same":
        downloads = doc_manager.get_doc_downloads(file_name)
        translation_files = find_translations(file_name, path, downloads)

    elif download_option == "folder" :
        downloads = doc_manager.get_doc_downloads(file_name)

        entry = doc_manager.get_doc_by_prop("file_name", file_name)
        if entry:
            file_name = entry['name']

        translation_files = find_translations(file_name, path, downloads)

    elif download_option == "clone":
        entry = doc_manager.get_doc_by_prop("file_name", file_name)
        if entry:
            file_name = entry['name']

        if os.path.isfile(os.path.join(path, file_name)):
            translation_files.append(os.path.join(path, file_name))

    return translation_files

def find_translations(file_name, path, downloads):
    translation_files = []
    trans_file_name = ""
    for d in downloads:
        temp = file_name.split(".")
        trans_file_name = ""
        for idx, val in enumerate(temp):
            if idx == len(temp)-2:
                trans_file_name = trans_file_name +val+"."
                trans_file_name = trans_file_name+d+"."
            else:
                trans_file_name += val
                if idx != len(temp)-1:
                    trans_file_name += "."

            if os.path.isfile(os.path.join(path, trans_file_name)):
                translation_files.append(os.path.join(path, trans_file_name))

    return translation_files


def raise_error(json, error_message, is_warning=False, doc_id=None, file_name=None):
    try:
        error = ""
        if json:
            error = json['messages'][0]
        if file_name:
            file_name = file_name.replace("Status of ", "")
        if file_name is not None and doc_id is not None:
            error = error.replace(doc_id, file_name+" ("+doc_id+")")
        # Sometimes api returns vague errors like 'Unknown error'
        if error == 'Unknown error':
            error = error_message
        if not is_warning:
            raise exceptions.RequestFailedError(error)
        # warnings.warn(error)
        if error:
            error = error+"\n"
        logger.error(error+error_message)
    except (AttributeError, IndexError):
        if not is_warning:
            raise exceptions.RequestFailedError(error_message)
        # warnings.warn(error_message)
        logger.error(error_message)

def error(error_message):
    logger.error(error_message+"\n")

def underline(text):
    if term:
        print(term.underline(text))
    else:
        # print("Recommended to install blessings module for better formatting")
        print(text)

def format_commit_message():
    pass

def check_response(response):
    try:
        if response and response.text:
            if response.json():
                return True
    except ValueError:
        logger.warning("Could not connect to Lingotek")
        return
    # Python 3
#     except json.decoder.JSONDecodeError:
#         logger.warning("Could not connect to Lingotek")
#         return
    # End Python 3

def remove_begin_slashes(path):
    index = 0
    for letter in path:
        if letter != os.sep:
            break
        index += 1
    if len(path) > index + 1:
        return path[index:]
    else:
        return ''
    return path


def remove_end_slashes(path):
    index = len(path) - 1
    for letter in reversed(path):
        if letter != os.sep:
            break
        index -= 1
    if index > 0:
        return path[:index - 1]
    else:
        return ''
    return path

def remove_last_folder_in_path(path):
    if len(path):
        split_path = path.split(os.sep)
        split_path = split_path[:len(split_path) - 1]
        return os.path.join(*split_path)
    else:
        return path

# Takes a path normalized relative to the project root (path) and returns the path relative to the current directory.
def get_relative_path(path_to_project_root, path):
    abs_path = os.path.dirname(os.path.join(path_to_project_root,path))
    # print("abs_path: "+abs_path)
    relative_path = os.path.relpath(abs_path,os.getcwd())
    # print("relative path: "+relative_path)
    if relative_path == '..' and os.path.join(path_to_project_root,path) == os.getcwd():
        return '.'
    relative_file_path = os.path.join(relative_path,os.path.basename(path))
    split_path = relative_file_path.split(os.sep)
    # print("cwd: "+os.getcwd())
    # print("joined path: "+os.path.join(abs_path,os.path.basename(path)))

    if len(split_path) and split_path[0] == '.' or os.path.join(abs_path,os.path.basename(path)) in os.getcwd():
        relative_file_path = os.path.join(*split_path[1:])
    return relative_file_path

def log_traceback(ex, ex_traceback=None):
    # Python 2
    try:
        exc_type, exc_value, exc_traceback = sys.exc_info()
        tb_str = ""
        tb_lines = traceback.format_exception(exc_type, exc_value, exc_traceback)
        for line in tb_lines:
            tb_str += line+"\n"
    finally:
       """Assigning the traceback return value to a local variable in a function that is handling an exception will cause a circular reference,
          preventing anything referenced by a local variable in the same function or by the traceback from being garbage collected.
          Must explicitly delete. """
       del exc_traceback
       return tb_str
    # End Python 2
    # Python 3
#     if ex_traceback is None:
#         ex_traceback = ex.__traceback__
#     tb_str = ""
#     tb_lines = traceback.format_exception(ex.__class__, ex, ex_traceback)
#     for line in tb_lines:
#         tb_str += line+"\n"
#     return tb_str
    # End Python 3

def log_error(error_file_name, e):
    try:
        with open(error_file_name, 'a') as error_file:
                error_file.write(str(time.strftime("%Y-%m-%d %H:%M:%S") + ": "+str(log_traceback(e))))
    except IOError as e:
        print(e.errno)
        print(e)
    return

def remove_powershell_formatting(args):
    if args != None:
        if isinstance(args, tuple):
            myTuple = ()
            if len(args) > 1:
                if isinstance(args, tuple):
                    for k in args:
                        k = remove_formatting(k)
                        myTuple = myTuple+(k,)

                    return myTuple
                else:
                    for k,v  in args:
                        k = (remove_formatting(k),)
                        v = remove_formatting(v)
                        tup1 = k+(v,)

                        return myTuple+(tup1,)
                return myTuple+(tup1,)
            else:
                for tup in args:
                    if isinstance(tup, tuple):
                        for k in tup:
                            k = remove_formatting(k)
                            myTuple = myTuple+(k,)

                        myTuple = (myTuple),
                        return myTuple
                    else:
                        for k in args:
                            k = remove_formatting(k)
                            myTuple = (k,)

                        return myTuple

                return args

        elif isinstance(args, list):
            temp = []
            for k in args:
                k = remove_formatting(k)
                temp.append(k)

            return tuple(temp)

        elif isinstance(args, str):
            temp = remove_formatting(args)
            return temp
        # Python 2
        elif isinstance(args, bool):
           return args
        # End Python 2
        else:
            # Python 2
          temp = remove_formatting(args)
          return temp
            # End Python 2
            # Python 3
#             return args
            # End Python 3

def remove_formatting(f):
    if f.startswith(".\\"):
        f = f[2:]

        if f.endswith("\\"):
            f = f[:-1]

        if f.endswith("\""):
            f = f[:-1]

        return f

    else:
        return f

def yes_no_prompt(prompt, default_yes=False):
    try:
        confirm = 'none'
        prompt_message = prompt + (" [Y/n]: " if default_yes else " [y/N]: ")
        while confirm not in ['y', 'Y', 'n', 'N', '']:
            # Python 2
            confirm = raw_input(prompt_message)
            # End Python 2
            # Python 3
#             confirm = input(prompt_message)
            # End Python 3
        if confirm in ['y', 'Y']:
            return True
        elif confirm in ['n', 'N']:
            return False
        else:
            return default_yes
    except KeyboardInterrupt:
        raise
